
# ruff: noqa: E501
# Imports
import os

import stouputils as stp
from beet.core.utils import JsonDict
from PIL import Image

from ...core.__memory__ import Mem
from ...core.cls.ingredients import FURNACES_RECIPES_TYPES, Ingr
from .image_utils import add_border, careful_resize, image_count
from .shared_import import BORDER_SIZE, SQUARE_SIZE, TEMPLATES_PATH, WIKI_INGR_OF_CRAFT_FONT, SharedMemory, get_border_color, get_next_font


# Generate page font function (called in utils)
def generate_page_font(name: str, page_font: str, craft: JsonDict|None = None, output_name: str = "") -> None:
	""" Generate the page font image with the proper items
	Args:
		name			(str):			Name of the item
		page_font		(str):			Font string for the page
		craft			(dict|None):	Crafting recipe dictionary (None if no craft)
		output_name		(str|None):		The output name (None if default, used for wiki crafts)
	"""
	# Skip if high resolution and craft is defined
	if SharedMemory.high_resolution and craft is not None:
		return

	# Determine output filename
	if not output_name:
		output_filename = name
	else:
		output_filename = output_name

	# Get result texture (to place later)
	image_path = f"{SharedMemory.cache_path}/items/{Mem.ctx.project_id}/{name}.png"
	if not os.path.exists(image_path):
		stp.warning(f"Missing texture at '{image_path}', using placeholder texture")
		result_texture = Image.new("RGBA", (SQUARE_SIZE, SQUARE_SIZE), (255, 255, 255, 0))  # Placeholder image
	else:
		result_texture = Image.open(image_path)

	# If recipe result is specified, take the right texture
	if craft and craft.get("result"):
		result_id = Ingr(craft["result"]).to_id()
		result_id = result_id.replace(":", "/")
		image_path = f"{SharedMemory.cache_path}/items/{result_id}.png"
		if not os.path.exists(image_path):
			stp.warning(f"Missing texture at '{image_path}', using placeholder texture")
			result_texture = Image.new("RGBA", (SQUARE_SIZE, SQUARE_SIZE), (255, 255, 255, 0))
		else:
			result_texture = Image.open(image_path)

	# Check if there is a craft and not high resolution
	if craft:

		# Resize the texture and get the mask
		result_texture = careful_resize(result_texture, SQUARE_SIZE)
		result_mask = result_texture.convert("RGBA").split()[3]

		# Shaped craft
		if craft["type"] in "crafting_shaped":
			shape: list[str] = craft["shape"]

			# Adjust craft shape for special cases
			# Special case: if it's 1 line with 3 columns, add empty lines to center it
			if len(shape) == 1 and len(shape[0]) == 3:
				shape = ["   ", shape[0], "   "]

			# Special case: if it's 3 lines with 1 column each, center them horizontally
			elif (len(shape) == 3 and
				all(len(shape_line) == 1 for shape_line in shape)):
				shape = [" " + line + " " for line in shape]

			# Get the image template and append the provider
			shaped_size = max(2, max(len(shape), len(shape[0])))
			template = Image.open(f"{TEMPLATES_PATH}/shaped_{shaped_size}x{shaped_size}.png")
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/page/{output_filename}.png", "ascent": 0 if not output_name else 6, "height": 60, "chars": [page_font]})

			# Loop the shape matrix
			STARTING_PIXEL = (4, 4)
			CASE_OFFSETS = (4, 4)
			for i, row in enumerate(shape):
				for j, symbol in enumerate(row):
					if symbol != " ":
						ingredient = Ingr(craft["ingredients"][symbol])
						if ingredient.get("components"):
							# get "iyc:steel_ingot" in {'components': {'custom_data': {'iyc': {'steel_ingot': True}}}}
							item = ingredient.to_id()
						else:
							item = ingredient["item"]	# Vanilla item, ex: "minecraft:glowstone"

						# Get the texture and place it at the coords
						item = item.replace(":", "/")
						image_path = f"{SharedMemory.cache_path}/items/{item}.png"
						if not os.path.exists(image_path):
							stp.warning(f"Missing texture at '{image_path}', using placeholder texture")
							item_texture = Image.new("RGBA", (SQUARE_SIZE, SQUARE_SIZE), (255, 255, 255, 0))  # Placeholder image
						else:
							item_texture = Image.open(image_path)
						item_texture = careful_resize(item_texture, SQUARE_SIZE)
						coords = (
							j * (SQUARE_SIZE + CASE_OFFSETS[0]) + STARTING_PIXEL[0],
							i * (SQUARE_SIZE + CASE_OFFSETS[1]) + STARTING_PIXEL[1]
						)
						mask = item_texture.convert("RGBA").split()[3]
						template.paste(item_texture, coords, mask)

			# Place the result item
			coords = (148, 40) if shaped_size == 3 else (118, 25)
			template.paste(result_texture, coords, result_mask)

			# Place count if the result is greater than 1
			if craft.get("result_count", 1) > 1:
				count_img = image_count(craft["result_count"])
				template.paste(count_img, [x + 2 for x in coords], count_img)	# type: ignore

			# Save the image
			template.save(f"{SharedMemory.cache_path}/font/page/{output_filename}.png")

		# Smelting craft
		elif craft["type"] in FURNACES_RECIPES_TYPES:

			# Get the image template and append the provider
			template = Image.open(f"{TEMPLATES_PATH}/furnace.png")
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/page/{output_filename}.png", "ascent": 0 if not output_name else 6, "height": 60, "chars": [page_font]})

			# Place input item
			input_item: str = Ingr(craft["ingredient"]).to_id()
			input_item = input_item.replace(":", "/")
			image_path = f"{SharedMemory.cache_path}/items/{input_item}.png"
			if not os.path.exists(image_path):
				stp.warning(f"Missing texture at '{image_path}', using placeholder texture")
				item_texture = Image.new("RGBA", (SQUARE_SIZE, SQUARE_SIZE), (255, 255, 255, 0))  # Placeholder image
			else:
				item_texture = Image.open(image_path)
			item_texture = careful_resize(item_texture, SQUARE_SIZE)
			mask = item_texture.convert("RGBA").split()[3]
			template.paste(item_texture, (4, 4), mask)

			# Place the result item and count if the result is greater than 1
			coords = (124, 40)
			template.paste(result_texture, coords, result_mask)
			if craft["result_count"] > 1:
				count_img = image_count(craft["result_count"])
				template.paste(count_img, [x + 2 for x in coords], count_img)	# type: ignore

			# Save the image
			template.save(f"{SharedMemory.cache_path}/font/page/{output_filename}.png")

	# Else, there is no craft, just put the item in a box
	else:
		# Get the image template and append the provider
		template = Image.open(f"{TEMPLATES_PATH}/simple_case_no_border.png")
		factor: int = 1
		if SharedMemory.high_resolution:
			factor_float: float = 256 / template.size[0]
			result_texture = careful_resize(result_texture, round(SQUARE_SIZE * factor_float))
			template = careful_resize(template, 256)
			result_mask = result_texture.convert("RGBA").split()[3]
			factor = int(factor_float)
		else:
			# Resize the texture and get the mask
			result_texture = careful_resize(result_texture, SQUARE_SIZE)
			result_mask = result_texture.convert("RGBA").split()[3]
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/page/{output_filename}.png", "ascent": 0 if not output_name else 6, "height": 40, "chars": [page_font]})

		# Place the result item
		template.paste(result_texture, (2 * factor, 2 * factor), result_mask)
		template = add_border(template, get_border_color(), BORDER_SIZE)
		template.save(f"{SharedMemory.cache_path}/font/page/{output_filename}.png")
	return


# Generate small craft icon
def generate_wiki_font_for_ingr(name: str, craft: JsonDict) -> str:
	""" Generate the wiki icon font to display in the manual for wiki buttons showing the result of the craft
	If no texture found for the resulting item, return the default wiki font
	Args:
		name	(str):		The name of the item, ex: "adamantium_fragment"
		craft	(JsonDict):	The associed craft, ex: {"type": "crafting_shaped","result_count": 1,"category": "equipment","shape": ["XXX","X X"],"ingredients": {"X": {"components": {"custom_data": {"iyc": {"adamantium_fragment": true}}}}},"result": {"components": {"custom_data": {"iyc": {"adamantium_helmet": true}}},"count": 1}}
	Returns:
		str: The craft icon
	"""
	# Default wiki font
	font = WIKI_INGR_OF_CRAFT_FONT

	# If no result found, return the default font
	if not craft.get("result"):
		return font

	# Get result item texture and paste it on the wiki_ingredient_of_craft_template
	try:
		craft_type = craft["type"]
		result_item: str = Ingr(craft["result"]).to_id().replace(":", "/")
		texture_path = f"{SharedMemory.cache_path}/items/{result_item}.png"
		result_item = result_item.replace("/", "_")
		dest_path = f"{SharedMemory.cache_path}/font/wiki_icons/{result_item}_{craft_type}.png"

		# Load texture and resize it
		item_texture = Image.open(texture_path)
		item_res = 64 if not SharedMemory.high_resolution else 256
		item_res_adjusted = int(item_res * 0.75)
		item_texture = careful_resize(item_texture, item_res_adjusted)
		item_texture = item_texture.convert("RGBA")

		# Load the template
		filename: str = "wiki_ingredient_of_craft_template.png" if craft_type != "mining" else "wiki_mining_template.png"
		template = Image.open(f"{TEMPLATES_PATH}/{filename}")
		template = careful_resize(template, item_res)

		# Paste the texture on it
		offset = (item_res - item_res_adjusted) // 2
		template.paste(item_texture, (offset, offset), item_texture)

		# Save the result
		template.save(dest_path)

		# Prepare provider
		font = get_next_font()
		rel_path: str = dest_path.replace(f"{SharedMemory.cache_path}/", f"{Mem.ctx.project_id}:")
		SharedMemory.font_providers.append({"type":"bitmap","file":rel_path, "ascent": 8, "height": 16, "chars": [font]})

	except Exception as e:
		stp.warning(f"Failed to generate craft icon for {name}: {e}\nreturning default font...")
		pass

	# Return the font
	return font




"""
Execution context parsing utilities for Minecraft function headers.

This module handles parsing execute commands to extract execution contexts
like 'as @e[...] & at @s' from command lines.
"""

# Imports
from beet.core.utils import JsonDict


# Functions
def parse_execution_context_from_line(line: str) -> str | None:
    """ Parse execution context from a line that calls a function.

    Args:
        line (str): The line containing the function call

    Returns:
        str | None: The execution context, or None if default
    """
    line = line.strip()

    # If it's not an execute command, no specific context
    if not line.startswith("execute "):
        return None

    # Dictionary mapping execute keywords to number of arguments they take
    execute_keywords: JsonDict = {
        "as": 1,
        "at": 1,
        "positioned": 3,
        "anchored": 1,
        "align": 1,
        "rotated": 2,
        "facing": None,  # Special case: variable arguments
        "in": 1,
    }

    def parse_selector_or_argument(parts: list[str], start_index: int) -> tuple[str, int]:
        """ Parse a selector that might contain square brackets or a simple argument.

        Args:
            parts (list[str]): The split command parts
            start_index (int): Index to start parsing from

        Returns:
            tuple[str, int]: The parsed argument and the next index to continue from
        """
        if start_index >= len(parts):
            return "", start_index

        arg: str = parts[start_index]
        next_index: int = start_index + 1

        # If not [, we assume it's a simple argument
        if "[" not in arg and arg.startswith("@"):
            return arg, next_index

        # If the argument contains [ but doesn't end with ], we need to collect more parts
        if "[" in arg and not arg.endswith("]"):
            # Keep collecting parts until we find one that ends with ]
            while next_index < len(parts) and not arg.endswith("]"):
                arg += " " + parts[next_index]
                next_index += 1

        # Clean up the selector by removing spaces after commas
        arg = arg.replace(", ", ",")

        # Simplify specific selector arguments if they exist
        if "[" in arg and "]" in arg:
            # Process duplicates and simplify only when multiple occurrences exist
            if "," in arg:
                # Extract the content between [ and ]
                selector_start: int = arg.find("[")
                selector_end: int = arg.rfind("]")
                if selector_start != -1 and selector_end != -1:
                    prefix: str = arg[:selector_start + 1]
                    suffix: str = arg[selector_end:]
                    content: str = arg[selector_start + 1:selector_end]

                    # Split by comma and process each part
                    parts_list: list[str] = [part.strip() for part in content.split(",")]

                    # Count occurrences of each attribute type
                    attribute_counts: dict[str, int] = {}
                    for part in parts_list:
                        if "=" in part:
                            attr_name: str = part.split("=")[0]
                            attribute_counts[attr_name] = attribute_counts.get(attr_name, 0) + 1

                    # Process parts and replace with ... only when there are duplicates
                    processed_parts: list[str] = []
                    seen_attributes: set[str] = set()

                    for part in parts_list:
                        if "=" in part:
                            attr_name: str = part.split("=")[0]
                            attr_value: str = part.split("=", 1)[1]  # Get the full value after the first =

                            # Special handling for NBT - only simplify if longer than 50 characters
                            if attr_name == "nbt":
                                if len(attr_value) > 50:
                                    if part.startswith(attr_name + "=!"):
                                        processed_parts.append(f"{attr_name}=!{{...}}")
                                    else:
                                        processed_parts.append(f"{attr_name}={{...}}")
                                else:
                                    processed_parts.append(part)  # Keep original if <= 50 chars
                            # If this attribute appears multiple times and we haven't processed it yet
                            elif attribute_counts[attr_name] > 1 and attr_name not in seen_attributes:
                                seen_attributes.add(attr_name)
                                # Replace with simplified version
                                if part.startswith(attr_name + "=!"):
                                    # Negative attribute
                                    if attr_name in ["tag", "predicate"]:
                                        processed_parts.append(f"{attr_name}=!...")
                                    else:
                                        processed_parts.append(part)  # Keep original for unknown types
                                else:
                                    # Positive attribute
                                    if attr_name in ["tag", "predicate"]:
                                        processed_parts.append(f"{attr_name}=...")
                                    else:
                                        processed_parts.append(part)  # Keep original for unknown types
                            elif attribute_counts[attr_name] == 1:
                                # Single occurrence, keep as-is (unless it's NBT and was already handled above)
                                if attr_name != "nbt":
                                    processed_parts.append(part)
                            # Skip subsequent occurrences of duplicate attributes (except NBT which is handled separately)
                        else:
                            # Keep non-attribute parts (like dx=0,dy=0,dz=0)
                            processed_parts.append(part)

                    # Rebuild the selector
                    arg = prefix + ",".join(processed_parts) + suffix

        return arg, next_index

    # Parse execute command components
    parts: list[str] = line.split()
    context_parts: list[str] = []

    i = 1  # Skip "execute"
    while i < len(parts):
        part = parts[i]

        if part == "run":
            # We've reached the end of the execute subcommands
            break
        elif part in execute_keywords:
            arg_count: int | None = execute_keywords[part]

            if part == "facing":
                # Special handling for facing command
                if i + 1 < len(parts) and parts[i + 1] == "entity":
                    if i + 3 < len(parts):
                        context_parts.append(f"facing entity {parts[i + 2]} {parts[i + 3]}")
                        i += 4
                    else:
                        i += 1
                else:
                    # facing coordinates
                    if i + 3 < len(parts):
                        context_parts.append(f"facing {parts[i + 1]} {parts[i + 2]} {parts[i + 3]}")
                        i += 4
                    else:
                        i += 1
            elif part == "positioned":
                # Special handling for positioned (can have 1 or 3 args)
                if i + 3 < len(parts):
                    context_parts.append(f"positioned {parts[i + 1]} {parts[i + 2]} {parts[i + 3]}")
                    i += 4
                elif i + 1 < len(parts):
                    # Handle "positioned ~" or selector cases
                    context_parts.append(f"positioned {parts[i + 1]}")
                    i += 2
                else:
                    i += 1
            elif part == "at":
                # Special handling for "at @s" - it resets position/rotation context
                if i + 1 < len(parts):
                    selector: str
                    next_i: int
                    selector, next_i = parse_selector_or_argument(parts, i + 1)
                    # Remove any previous position/rotation modifiers when "at" is used
                    context_parts = [cp for cp in context_parts if not any(keyword in cp for keyword in ["positioned", "align", "rotated", "anchored", "in"])]
                    context_parts.append(f"{part} {selector}")
                    i = next_i
                else:
                    i += 1
            elif part == "as":
                # Special handling for selectors that might contain square brackets
                if i + 1 < len(parts):
                    selector, next_i = parse_selector_or_argument(parts, i + 1)
                    context_parts.append(f"{part} {selector}")
                    i = next_i
                else:
                    i += 1
            else:
                # Standard handling for keywords with fixed argument count
                if i + arg_count < len(parts):
                    args: str = " ".join(parts[i + 1:i + 1 + arg_count])
                    context_parts.append(f"{part} {args}")
                    i += 1 + arg_count
                else:
                    i += 1
        else:
            # Unknown keyword, skip it
            i += 1

    if context_parts:
        return " & ".join(context_parts)
    return None




# Imports
import json

import stouputils as stp
from beet.core.utils import JsonDict

from ...core.__memory__ import Mem
from ...core.cls.ingredients import Ingr
from ...core.cls.item import Item
from ...core.cls.recipe import CraftingShapedRecipe, CraftingShapelessRecipe
from ...core.constants import OFFICIAL_LIBS, official_lib_used
from ...core.utils.io import write_function


class SmithedRecipeHandler:
    """ Handler for Smithed Crafter recipe generation.

    This class handles the generation of custom recipes using Smithed Crafter.
    """

    def __init__(self) -> None:
        """ Initialize the handler. """
        self.SMITHED_SHAPELESS_PATH: str = f"{Mem.ctx.project_id}:calls/smithed_crafter/shapeless_recipes"
        self.SMITHED_SHAPED_PATH: str = f"{Mem.ctx.project_id}:calls/smithed_crafter/shaped_recipes"
        self.SMITHED_APPLY_PATH: str = f"{Mem.ctx.project_id}:calls/smithed_crafter/apply_recipe"

    @classmethod
    def routine(cls) -> None:
        """Main routine for Smithed Crafter recipe generation."""
        handler = cls()
        handler.generate_recipes()

    @stp.simple_cache
    def smithed_shapeless_recipe(self, recipe: CraftingShapelessRecipe, result_loot: str) -> str:
        """ Generate a Smithed Crafter shapeless recipe.

        Args:
            recipe (CraftingShapelessRecipe): The recipe data.
            result_loot (str): The loot table for the result.

        Returns:
            str: The generated recipe command.
        """
        # Get unique ingredients and their count
        unique_ingredients: list[tuple[int, JsonDict]] = []
        for ingr in recipe.ingredients:
            index: int = -1
            for i, (_, e) in enumerate(unique_ingredients):
                if str(ingr) == str(e):
                    index = i
                    break
            if index == -1:
                unique_ingredients.append((1, ingr))
            else:
                unique_ingredients[index] = (unique_ingredients[index][0] + 1, unique_ingredients[index][1])

        # Write the line
        line: str = (
            "execute if score @s smithed.data matches 0 store result score @s smithed.data "
            f"if score count smithed.data matches {len(unique_ingredients)} if data storage smithed.crafter:input "
        )
        r: dict[str, list[JsonDict]] = {"recipe": []}
        for count, ingr in unique_ingredients:
            item: JsonDict = {"count": count}
            item.update(ingr)
            r["recipe"].append(Ingr(item).item_to_id())
        line += json.dumps(r)

        if recipe.smithed_crafter_command:
            line += f""" run function {self.SMITHED_APPLY_PATH} {{"command":"{recipe.smithed_crafter_command}"}}"""
        else:
            line += f""" run function {self.SMITHED_APPLY_PATH} {{"command":"loot replace block ~ ~ ~ container.16 loot {result_loot}"}}"""
        return line

    @stp.simple_cache
    def smithed_shaped_recipe(self, recipe: CraftingShapedRecipe, result_loot: str) -> str:
        """ Generate a Smithed Crafter shaped recipe.

        Args:
            recipe (CraftingShapedRecipe): The recipe data.
            result_loot (str): The loot table for the result.

        Returns:
            str: The generated recipe command.
        """
        # Convert ingredients to aimed recipes
        ingredients: dict[str, Ingr] = recipe.ingredients
        recipes: dict[int, list[JsonDict]] = {0: [], 1: [], 2: []}

        for i, row in enumerate(recipe.shape):
            for slot, char in enumerate(row):
                ingredient = ingredients.get(char)
                if ingredient:
                    ingr = {"Slot": slot}
                    ingr.update(ingredient)
                    recipes[i].append(Ingr(ingr).item_to_id())
                else:
                    recipes[i].append({"Slot": slot, "id": "minecraft:air"})

        # Initialize the dump string
        dump: str = "{"

        # Iterate through each layer and its ingredients
        for i in range(3):
            if (i not in recipes) or (all(ingr.get("id") == "minecraft:air" for ingr in recipes[i])):
                recipes[i] = []

        for layer, ingrs in recipes.items():
            # If the list is empty, continue
            if not ingrs:
                dump += f"{layer}:[],"
                continue

            dump += f"{layer}:["  # Start of layer definition

            # Ensure each layer has exactly 3 ingredients by adding missing slots
            for i in range(len(ingrs), 3):
                ingrs.append({"Slot": i, "id": "minecraft:air"})

            # Process each ingredient in the layer
            for ingr in ingrs:
                ingr = ingr.copy()  # Create a copy to modify
                slot: int = ingr.pop("Slot")  # Extract the slot number
                ingr = json.dumps(ingr)[1:-1]  # Convert to JSON string without brackets
                dump += f'{{"Slot":{slot}b, {ingr}}},'  # Add the ingredient to the dump with its slot

            # Remove the trailing comma if present
            if dump[-1] == ',':
                dump = dump[:-1] + "],"  # End of layer definition
            else:
                dump += "],"  # End of layer definition without trailing comma

        # Remove the trailing comma if present and close the dump string
        if dump[-1] == ',':
            dump = dump[:-1] + "}"  # Close the dump string
        else:
            dump += "}"  # Close the dump string without trailing comma

        # Return the line
        line = f"execute if score @s smithed.data matches 0 store result score @s smithed.data if data storage smithed.crafter:input recipe{dump}"
        if recipe.smithed_crafter_command:
            line += f""" run function {self.SMITHED_APPLY_PATH} {{"command":"{recipe.smithed_crafter_command}"}}"""
        else:
            line += f""" run function {self.SMITHED_APPLY_PATH} {{"command":"loot replace block ~ ~ ~ container.16 loot {result_loot}"}}"""
        return line

    def generate_recipes(self) -> None:
        """ Generate all Smithed Crafter recipes. """
        for item in Mem.definitions.keys():
            obj = Item.from_id(item)

            for recipe in obj.recipes:
                if recipe["type"] not in (CraftingShapelessRecipe.type, CraftingShapedRecipe.type):
                    continue
                recipe = CraftingShapedRecipe.from_dict(recipe) \
                    if recipe["type"] == CraftingShapedRecipe.type \
                    else CraftingShapelessRecipe.from_dict(recipe)

                # Get ingredients
                ingr: list[Ingr] = list(recipe.ingredients.values()) if isinstance(recipe, CraftingShapedRecipe) else recipe.ingredients
                if not recipe.result:
                    result_loot_table = Ingr(item).register_loot_table(recipe.result_count)
                else:
                    result_loot_table = recipe.result.register_loot_table(recipe.result_count)

                # If there is a component in the ingredients of shaped/shapeless, use smithed crafter
                if any(i.get("components") for i in ingr):
                    if not official_lib_used("smithed.crafter"):
                        stp.debug("Found a crafting table recipe using custom item in ingredients, adding 'smithed.crafter' dependency")

                        # Add to the give_all function the heavy workbench give command
                        write_function(f"{Mem.ctx.project_id}:_give_all", "loot give @s loot smithed.crafter:blocks/table\n", prepend=True)

                # Generate recipe based on type
                if isinstance(recipe, CraftingShapelessRecipe):
                    line = self.smithed_shapeless_recipe(recipe, result_loot_table)
                    write_function(self.SMITHED_SHAPELESS_PATH, line, tags=["smithed.crafter:event/shapeless_recipes"])
                else:
                    line = self.smithed_shaped_recipe(recipe, result_loot_table)
                    write_function(self.SMITHED_SHAPED_PATH, line, tags=["smithed.crafter:event/recipes"])

        # Apply recipe
        if OFFICIAL_LIBS["smithed.crafter"]["is_used"]:
            write_function(self.SMITHED_APPLY_PATH, """
# Set the consume_tools flag
data modify storage smithed.crafter:input flags set value ["consume_tools"]

# Perform the loot command
$return run $(command)
""")





# ruff: noqa: E501
# Imports
import stouputils as stp
from beet import Context
from beet.core.utils import JsonDict

from ....core.__memory__ import Mem
from ....core.cls.item import Item
from ....core.utils.io import write_function_tag, write_load_file, write_versioned_function


# Main entry point
@stp.measure_time(message="Execution time of 'stewbeet.plugins.datapack.loading'")
def beet_default(ctx: Context):
	""" Main entry point for the datapack loading plugin.

	Requires plugin `stewbeet.plugins.finalyze.dependencies` later in the pipeline to complete.

	Args:
		ctx (Context): The beet context.
	"""
	# Set up memory context
	if Mem.ctx is None: # pyright: ignore[reportUnnecessaryComparison]
		Mem.ctx = ctx

	# Assertions
	assert ctx.project_version, "Project version is not set. Please set it in the project configuration."
	assert ctx.project_id, "Project ID is not set. Please set it in the project configuration."
	assert ctx.project_version.count(".") == 2, "Project version must be in the format 'major.minor.patch'."

	# Get basic project information
	major, minor, patch = ctx.project_version.split(".")

	# Setup enumerate and resolve functions
	write_versioned_function("load/enumerate",
f"""
# If current major is too low, set it to the current major
execute unless score #{ctx.project_id}.major load.status matches {major}.. run scoreboard players set #{ctx.project_id}.major load.status {major}

# If current minor is too low, set it to the current minor (only if major is correct)
execute if score #{ctx.project_id}.major load.status matches {major} unless score #{ctx.project_id}.minor load.status matches {minor}.. run scoreboard players set #{ctx.project_id}.minor load.status {minor}

# If current patch is too low, set it to the current patch (only if major and minor are correct)
execute if score #{ctx.project_id}.major load.status matches {major} if score #{ctx.project_id}.minor load.status matches {minor} unless score #{ctx.project_id}.patch load.status matches {patch}.. run scoreboard players set #{ctx.project_id}.patch load.status {patch}
""")

	write_versioned_function("load/resolve",
f"""
# If correct version, load the datapack
execute if score #{ctx.project_id}.major load.status matches {major} if score #{ctx.project_id}.minor load.status matches {minor} if score #{ctx.project_id}.patch load.status matches {patch} run function {ctx.project_id}:v{ctx.project_version}/load/main
""")

	# Setup enumerate and resolve function tags
	write_function_tag(f"{ctx.project_id}:enumerate", [f"{ctx.project_id}:v{ctx.project_version}/load/enumerate"])
	write_function_tag(f"{ctx.project_id}:resolve", [f"{ctx.project_id}:v{ctx.project_version}/load/resolve"])

	# Setup load main function
	write_versioned_function("load/main",
f"""
# Avoiding multiple executions of the same load function
execute unless score #{ctx.project_id}.loaded load.status matches 1 run function {ctx.project_id}:v{ctx.project_version}/load/secondary

""")

	# Confirm load
	items_storage = ""	# Storage representation of every item in the definitions
	if Mem.definitions:
		items_storage += f"\n# Items storage\ndata modify storage {ctx.project_id}:items all set value {{}}\n"
		for item in Mem.definitions.keys():
			obj = Item.from_id(item)

			# Prepare storage data with item_model component in first
			mc_data: JsonDict = {"id": obj.base_item, "count": 1, "components": {"minecraft:item_model": ""}}
			for k, v in obj.components.items():

				# Add 'minecraft:' if missing
				if ":" not in k:
					if k.startswith("!"):
						k = f"!minecraft:{k[1:]}"
					else:
						k = f"minecraft:{k}"

				# Copy component
				mc_data["components"][k] = v

			# If no item_model, remove it
			if mc_data["components"]["minecraft:item_model"] == "":
				del mc_data["components"]["minecraft:item_model"]

			# Append to the storage definitions, json_dump adds
			items_storage += f"data modify storage {ctx.project_id}:items all.{item} set value " + stp.json_dump(mc_data, max_level = 0)

	# Write the loading tellraw and score, along with the final dataset
	project_name = ctx.project_name or ctx.project_id
	write_load_file(
f"""
# Confirm load
tellraw @a[tag=convention.debug] {{"text":"[Loaded {project_name} v{ctx.project_version}]","color":"green"}}
scoreboard players set #{ctx.project_id}.loaded load.status 1
""" + items_storage)





# Imports
import os

import stouputils as stp
from beet import Context, PaintingVariant, PaintingVariantTag
from beet.core.utils import JsonDict

from ...core.__memory__ import Mem
from ...core.constants import PAINTING_DATA
from ...core.utils.io import set_json_encoder, texture_mcmeta


# Main entry point
@stp.measure_time(message="Execution time of 'stewbeet.plugins.custom_paintings'")
def beet_default(ctx: Context) -> None:
    """ Main entry point for the custom paintings plugin.
    This plugin handles the generation of custom paintings for the datapack and resource pack.

    Requires a valid definitions in Mem.definitions in order to function properly using constant 'PAINTING_DATA'.

    Args:
        ctx (Context): The beet context.
    """
    if Mem.ctx is None: # pyright: ignore[reportUnnecessaryComparison]
        Mem.ctx = ctx
    ns: str = ctx.project_id
    textures_folder: str = Mem.ctx.meta.get("stewbeet", {}).get("textures_folder", "")
    placeable_values: list[str] = []

    # Assertions
    assert textures_folder, "The 'textures_folder' key is missing in the 'stewbeet' section of the beet.yml file."

    # For each item definition that has painting data,
    for item, data in Mem.definitions.items():
        painting_data: JsonDict = data.get(PAINTING_DATA, {})
        if painting_data:

            ## Datapack
            # Add the item id to the list of painting variants values
            if not painting_data.get("not_placeable", False):
                placeable_values.append(f"{ns}:{item}")

            # Set default author and title if not provided
            if "author" not in painting_data:
                painting_data["author"] = {"text": Mem.ctx.project_author or "Unknown"}
            if "title" not in painting_data:
                painting_data["title"] = data.get("item_name") or {"text": item.replace("_", " ").title()}

            # Create ordered painting data with asset_id first
            texture_name: str = painting_data.get("texture", item)
            ordered_painting_data = {"asset_id": f"{ns}:{texture_name}"}
            ordered_painting_data.update(painting_data)
            ordered_painting_data.pop("not_placeable", None)
            ordered_painting_data.pop("texture", None)

            # Create the painting definition
            Mem.ctx.data[ns].painting_variants[item] = set_json_encoder(PaintingVariant(ordered_painting_data))

            ## Resource pack
            # Get the texture path
            if "texture" in painting_data:
                texture: str = painting_data["texture"]
                src: str = stp.relative_path(f"{textures_folder}/{texture}.png")
            else:
                matching_textures: list[str] = [
                    stp.relative_path(f"{root}/{file}")
                    for root, _, files in os.walk(textures_folder)
                    for file in files if file == f"{item}.png"
                ]
                if not matching_textures:
                    stp.error(f"No texture found for painting '{item}' in the textures folder '{textures_folder}'. Expected a file named '{item}.png'.")
                    continue
                elif len(matching_textures) > 1:
                    stp.warning(f"Multiple textures found for painting '{item}' in the textures folder '{textures_folder}'. Using the first one found: '{matching_textures[0]}'.")
                src: str = matching_textures[0]
            dst: str = f"painting/{texture_name}"

            # Check if the texture is not already registered
            if not Mem.ctx.assets[ns].textures.get(dst):
                Mem.ctx.assets[ns].textures[dst] = texture_mcmeta(src)

    # Add the painting variant tag to the context data
    if placeable_values:
        Mem.ctx.data["minecraft"].painting_variant_tags["placeable"] = set_json_encoder(PaintingVariantTag({"values": placeable_values}))





# Imports
import json

import stouputils as stp

from ...core.__memory__ import Mem
from ...core.cls.ingredients import Ingr
from ...core.cls.item import Item
from ...core.cls.recipe import PulverizingRecipe
from ...core.utils.io import write_function


class PulverizerRecipeHandler:
    """ Handler for pulverizer recipe generation.

    This class handles the generation of pulverizer recipes.
    """
    def __init__(self) -> None:
        """ Initialize the handler. """
        self.SIMPLENERGY_PULVERIZER_PATH: str = f"{Mem.ctx.project_id}:calls/simplenergy/pulverizer_recipes"

    @classmethod
    def routine(cls) -> None:
        """ Main routine for pulverizer recipe generation. """
        handler = cls()
        handler.generate_recipes()

    @stp.simple_cache
    def simplenergy_pulverizer_recipe(self, recipe: PulverizingRecipe, item: str) -> str:
        """ Generate a pulverizer recipe.

        Args:
            recipe (PulverizingRecipe): The recipe data.
            item (str): The item to generate the recipe for.

        Returns:
            str: The generated recipe command.
        """
        if not recipe.ingredient:
            recipe.ingredient = Ingr(item)
        ingredient = recipe.ingredient.item_to_id()
        result = recipe.result.to_item().item_to_id() if recipe.result else Ingr(item)

        line: str = "execute if score #found simplenergy.data matches 0 store result score #found simplenergy.data if data storage simplenergy:main pulverizer.input"
        line += json.dumps(ingredient)
        line += f" run loot replace entity @s contents loot {result.register_loot_table(recipe.result_count)}"
        return line

    def generate_recipes(self) -> None:
        """ Generate all pulverizer recipes. """
        for item in Mem.definitions.keys():
            obj = Item.from_id(item)

            for recipe in obj.recipes:
                if recipe["type"] == PulverizingRecipe.type:
                    if not recipe.get("ingredient"):
                        recipe["ingredient"] = Ingr(item)
                    recipe = PulverizingRecipe.from_dict(recipe)
                    write_function(
                        self.SIMPLENERGY_PULVERIZER_PATH,
                        self.simplenergy_pulverizer_recipe(recipe, item),
                        tags=["simplenergy:calls/pulverizer_recipes"],
                    )





# ruff: noqa: E501, RUF012
# Imports
from beet.core.utils import JsonDict
from beet.library import base
from beet.toolchain.config import FormatSpecifier

from ..dependencies.bookshelf import BOOKSHELF_MODULES

# Minecraft version specific constants
MORE_DATA_PACK_FORMATS: dict[tuple[int, ...], FormatSpecifier] = {
	(1, 21): 48,
	(1, 21, 0): 48,
	(1, 21, 1): 48,
	(1, 21, 2): 57,
	(1, 21, 3): 57,
	(1, 21, 4): 61,
	(1, 21, 5): 71,
	(1, 21, 6): 80,
	(1, 21, 7): 81,
	(1, 21, 8): 81,
	(1, 21, 9): (88, 0),
	(1, 21, 10): (88, 0),
	(1, 21, 11): (94, 1),

	# New version numbering system
	(25, 1): 71,			 # 1.21.5
	(25, 1, 0): 71,			 # 1.21.5
	(25, 2): 80,			# 1.21.6
	(25, 2, 0): 80,			# 1.21.6
	(25, 2, 1): 80,			# 1.21.6
	(25, 2, 2): 80,			# 1.21.6
	(25, 3): 81,			 # 1.21.8
	(25, 3, 0): 81,			 # 1.21.8
	(25, 3, 1): 81,			 # 1.21.8
	(25, 4): (94, 1),		# 1.21.11
	(25, 4, 0): (94, 1),	# 1.21.11

	(9999, 0, 0): (9999, 0),
}
MORE_ASSETS_PACK_FORMATS: dict[tuple[int, ...], FormatSpecifier] = {
	(1, 21): 34,
	(1, 21, 0): 34,
	(1, 21, 1): 34,
	(1, 21, 2): 42,
	(1, 21, 3): 42,
	(1, 21, 4): 46,
	(1, 21, 5): 55,
	(1, 21, 6): 63,
	(1, 21, 7): 64,
	(1, 21, 8): 64,
	(1, 21, 9): (69, 0),
	(1, 21, 10): (69, 0),
	(1, 21, 11): (75, 0),

	# New version numbering system
	(25, 1): 55,			 # 1.21.5
	(25, 1, 0): 55,			 # 1.21.5
	(25, 2): 63,			 # 1.21.6
	(25, 2, 0): 63,			# 1.21.6
	(25, 2, 1): 63,			# 1.21.6
	(25, 2, 2): 63,			# 1.21.6
	(25, 3): 64,			 # 1.21.8
	(25, 3, 0): 64,			 # 1.21.8
	(25, 3, 1): 64,			 # 1.21.8
	(25, 4): (75, 0),		# 1.21.11
	(25, 4, 0): (75, 0),	# 1.21.11

	(9999, 0, 0): (9999, 0),
}
MORE_DATA_VERSIONS: dict[tuple[int, ...], int] = {
	(1, 21): 3953,
	(1, 21, 0): 3953,
	(1, 21, 1): 3955,
	(1, 21, 2): 4080,
	(1, 21, 3): 4082,
	(1, 21, 4): 4189,
	(1, 21, 5): 4325,
	(1, 21, 6): 4435,
	(1, 21, 7): 4438,
	(1, 21, 8): 4440,
	(1, 21, 9): 4554,
	(1, 21, 10): 4556,
	(1, 21, 11): 4669,
}
LATEST_MC_VERSION: str = ".".join(str(x) for x in list(MORE_DATA_VERSIONS.keys())[-1])
base.LATEST_MINECRAFT_VERSION = LATEST_MC_VERSION

# Databases
CATEGORY: str = "manual_category"						# Key for the category, used for recipes and the manual, ex: CATEGORY:"material" or CATEGORY:"equipment"
CUSTOM_BLOCK_VANILLA: str = "minecraft:furnace"			# Vanilla block used as base for custom blocks, must have the "facing" blockstate
CUSTOM_BLOCK_ALTERNATIVE: str = "minecraft:item_frame"	# Same purpose as previous, but useful for blocks that can be placed on walls or on player's position (ex: flowers)
CUSTOM_BLOCK_HEAD: str = "minecraft:player_head"		# Same purpose as previous, but useful for blocks does not have a custom model data
CUSTOM_ITEM_VANILLA: str = "minecraft:command_block"	# Vanilla item used as base for custom items, must not have any survival vanilla behaviour
VANILLA_BLOCK: str = "vanilla_block"					# Key to a vanilla block that will be placed for custom block interaction, value needs to be a dict like {"id":"minecraft:chest[type=single,waterlogged=false]", "apply_facing": True}
NO_SILK_TOUCH_DROP: str = "no_silk_touch_drop"			# Key to an item ID that will drop when silk touch is not used. Must be used only when using the vanilla block for ores, ex: "adamantium_fragment" or "minecraft:raw_iron"
OVERRIDE_MODEL: str = "override_model"					# Key to a dictionnary that will be used to override the whole model
SMITHED_CRAFTER_COMMAND: str = "smithed_crafter_command"	# Key to a command that will be used in a recipe in the Smithed Crafter library. If not present, the command will be defaulted to a loot table. Ex: {"result":...,SMITHED_CRAFTER_COMMAND: "function your_namespace:calls/smithed_crafter/do_something_else"}
PAINTING_DATA: str = "painting_data"					# Key to a dict that contains the painting data, like {"author":"","title":"","width":1,"height":1} where author and title defaults to beet config values if not given
GROWING_SEED: str = "growing_seed"						# Key to a seed that has multiple growth stages, value needs to be a dict like {"texture_basename":"wheat","stages":8,"seconds":600,"planted_on":"stone","loots":[{"id":"minecraft:wheat_seeds","min_count":1,"max_count":3}]} where loots can be either this format or a loot table path "namespace:blocks/loot_table_name"
WIKI_COMPONENT: str = "wiki_buttons"					# Key to a text component that will be used to generate the wiki button in the manual
RESULT_OF_CRAFTING: str = "result_of_crafting"			# Key to a list of recipes to craft the item, ex: "adamantium": {RESULT_OF_CRAFTING: [...]}
USED_FOR_CRAFTING: str = "used_for_crafting"			# Should not be used unless you are crafting a vanilla item (ex: iyc.chainmail -> chainmail armor)
NOT_COMPONENTS: list[str] = [							# Keys that should not be considered as components. Used for recipes, loot tables, etc.
	"id",
	WIKI_COMPONENT,
	RESULT_OF_CRAFTING,
	USED_FOR_CRAFTING,
	CATEGORY,
	VANILLA_BLOCK,
	NO_SILK_TOUCH_DROP,
	OVERRIDE_MODEL,
	SMITHED_CRAFTER_COMMAND,
	PAINTING_DATA,
	GROWING_SEED,
]

# Technical constants
COMMON_SIGNAL: str = r'custom_data={"common_signals":{"temp":true}}'					# NBT to add to items to mark them as temporary for Common Signals library
COMMON_SIGNAL_HIDDEN: str = r'tooltip_display={"hide_tooltip":true},' + COMMON_SIGNAL	# Same as previous but also hide the tooltip (Useful for GUIs)
FACES: tuple[str, ...] = ("down", "up", "north", "south", "west", "east")						# Faces of a block, used for resource pack and blocks orientation
SIDES: tuple[str, ...] = ("_bottom", "_top", "_front", "_back", "_left", "_right", "_side")	# Sides of a block, used for resource pack
DOWNLOAD_VANILLA_ASSETS_RAW = "https://raw.githubusercontent.com/edayot/renders/renders/resourcepack/assets/minecraft/textures/render"
DOWNLOAD_VANILLA_ASSETS_SPECIAL_RAW = "https://raw.githubusercontent.com/edayot/renders/renders-special/resourcepack/assets/minecraft/textures/render"
DOWNLOAD_VANILLA_ASSETS_SOURCE = "https://github.com/edayot/renders/tree/renders/resourcepack/assets/minecraft/textures/render"
CUSTOM_BLOCK_HEAD_CUBE_RADIUS: tuple[int, int, int] = (16, 16, 16)	# Size of the region to check around the player when placing a CUSTOM_BLOCK_HEAD
BLOCKS_WITH_INTERFACES: list[str] = [	# List of blocks that are containers and have an interface
	"minecraft:barrel",
	"minecraft:chest",
	"minecraft:furnace",
	"minecraft:shulker_box",
	"minecraft:dispenser",
	"minecraft:hopper",
	"minecraft:dropper",
	"minecraft:smoker",
	"minecraft:blast_furnace",
	"minecraft:brewing_stand",
	"minecraft:trapped_chest",
	"minecraft:crafter",
]

# Conventions constants
class Conventions:
	""" Defines conventions for tags used in datapacks. """
	NO_KILL_TAGS: list[str] = ["smithed.strict", "global.ignore.kill"]
	""" List of tags that prevent entities from being killed. """
	ENTITY_TAGS: list[str] = ["smithed.entity", "global.ignore"]
	""" List of tags applicable to custom entities. """
	BLOCK_TAGS: list[str] = ["smithed.block", *ENTITY_TAGS]
	""" List of tags applicable to custom blocks. """
	ENTITY_TAGS_NO_KILL: list[str] = ENTITY_TAGS + NO_KILL_TAGS
	""" Combined list of entity tags and no kill tags. """
	BLOCK_TAGS_NO_KILL: list[str] = BLOCK_TAGS + NO_KILL_TAGS
	""" Combined list of block tags and no kill tags. """

	AVOID_NO_KILL: str = ",".join(f"tag=!{tag}" for tag in NO_KILL_TAGS)
	""" String of tags to avoid when killing entities. Example of use: execute as @e[{Conventions.AVOID_NO_KILL}] run function your_namespace:kill_entity """
	AVOID_ENTITY_TAGS: str = ",".join(f"tag=!{tag}" for tag in ENTITY_TAGS)
	""" String of tags to avoid when executing an entity command. Example of use: execute as @e[{Conventions.AVOID_ENTITY_TAGS}] run function your_namespace:kill_entity """
	AVOID_BLOCK_TAGS: str = ",".join(f"tag=!{tag}" for tag in BLOCK_TAGS)
	""" String of tags to avoid when executing a block command. Example of use: execute as @e[{Conventions.AVOID_BLOCK_TAGS}] run function your_namespace:kill_entity """
	AVOID_ENTITY_TAGS_NO_KILL: str = ",".join(f"tag=!{tag}" for tag in ENTITY_TAGS_NO_KILL)
	""" String of tags to avoid when executing an entity command. Example of use: execute as @e[{Conventions.AVOID_ENTITY_TAGS_NO_KILL}] run function your_namespace:kill_entity """
	AVOID_BLOCK_TAGS_NO_KILL: str = ",".join(f"tag=!{tag}" for tag in BLOCK_TAGS_NO_KILL)
	""" String of tags to avoid when executing a block command. Example of use: execute as @e[{Conventions.AVOID_BLOCK_TAGS_NO_KILL}] run function your_namespace:kill_entity """



# Automatically handled dependencies for supported libs with additional key "is_used" that is True when the lib is found to be used.
def official_lib_used(lib: str) -> bool:
	is_used: bool = OFFICIAL_LIBS[lib]["is_used"]
	OFFICIAL_LIBS[lib]["is_used"] = True
	return is_used

OFFICIAL_LIBS: dict[str, JsonDict] = {
	"common_signals":		{"version":[0, 2, 0],	"name":"Common Signals",					"url":"https://github.com/Stoupy51/CommonSignals",			"is_used": False},
	"smithed.custom_block":	{"version":[0, 7, 1],	"name":"Smithed Custom Block",				"url":"https://wiki.smithed.dev/libraries/custom-block/",	"is_used": False},
	"smithed.crafter":		{"version":[0, 7, 1],	"name":"Smithed Crafter",					"url":"https://wiki.smithed.dev/libraries/crafter/",		"is_used": False},
	"furnace_nbt_recipes":	{"version":[1, 10, 1],	"name":"Furnace NBT Recipes",				"url":"https://github.com/Stoupy51/FurnaceNbtRecipes",		"is_used": False},
	"smart_ore_generation":	{"version":[1, 7, 2],	"name":"SmartOreGeneration",				"url":"https://github.com/Stoupy51/SmartOreGeneration",		"is_used": False},
	"itemio":				{"version":[1, 4, 1],	"name":"ItemIO",							"url":"https://github.com/edayot/ItemIO",					"is_used": False},
	**BOOKSHELF_MODULES,
}





# Imports
from beet.core.utils import JsonDict

# This file is auto-generated by the release script (python_package/scripts/on_bookshelf_release.py).
# Do not edit this file manually.
BOOKSHELF_MODULES: dict[str, JsonDict] = {
	"bs.bitwise": {"version": [3,2,0],"name": "Bookshelf Bitwise","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.block": {"version": [3,2,0],"name": "Bookshelf Block","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.color": {"version": [3,2,0],"name": "Bookshelf Color","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.dump": {"version": [3,2,0],"name": "Bookshelf Dump","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.environment": {"version": [3,2,0],"name": "Bookshelf Environment","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.generation": {"version": [3,2,0],"name": "Bookshelf Generation","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.health": {"version": [3,2,0],"name": "Bookshelf Health","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.hitbox": {"version": [3,2,0],"name": "Bookshelf Hitbox","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.id": {"version": [3,2,0],"name": "Bookshelf Id","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.interaction": {"version": [3,2,0],"name": "Bookshelf Interaction","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.link": {"version": [3,2,0],"name": "Bookshelf Link","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.log": {"version": [3,2,0],"name": "Bookshelf Log","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.math": {"version": [3,2,0],"name": "Bookshelf Math","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.move": {"version": [3,2,0],"name": "Bookshelf Move","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.position": {"version": [3,2,0],"name": "Bookshelf Position","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.random": {"version": [3,2,0],"name": "Bookshelf Random","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.raycast": {"version": [3,2,0],"name": "Bookshelf Raycast","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.schedule": {"version": [3,2,0],"name": "Bookshelf Schedule","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.sidebar": {"version": [3,2,0],"name": "Bookshelf Sidebar","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.spline": {"version": [3,2,0],"name": "Bookshelf Spline","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.string": {"version": [3,2,0],"name": "Bookshelf String","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.time": {"version": [3,2,0],"name": "Bookshelf Time","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.tree": {"version": [3,2,0],"name": "Bookshelf Tree","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.vector": {"version": [3,2,0],"name": "Bookshelf Vector","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.view": {"version": [3,2,0],"name": "Bookshelf View","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False},
	"bs.xp": {"version": [3,2,0],"name": "Bookshelf Xp","url": "https://github.com/mcbookshelf/bookshelf/releases","is_used": False}
}




from _typeshed import Incomplete
from beet import Context as Context

GENERATION_PLUGINS: Incomplete
FINALYZE_PLUGINS: Incomplete

def beet_default(ctx: Context):
    """ Run all plugins that should run after definitions are loaded.
    This function will yield before finalizing the project (so users can run their own plugins in the middle of the pipeline).

    Requires:
        - StewBeet to be initialized (e.g. from stewbeet.plugins.initialize).
        - Definitions to be loaded
    """




# Imports
from ...core import Item, Mem, write_function, write_load_file


# Add commands to place and destroy functions for energy items
def insert_lib_calls() -> None:
	ns: str = Mem.ctx.project_id

	# Create new energy_rate scoreboard objective
	write_load_file(f"\n# Score for energy usage or generation\nscoreboard objectives add {ns}.energy_rate dummy\n", prepend = True)

	# Loop through all items with energy data
	for item, data in Mem.definitions.items():
		obj = Item.from_dict(data, item)
		energy: dict[str, int] = obj.components.get("custom_data", {}).get("energy", {})
		if len(energy) > 0:
			placement: str = f"{ns}:custom_blocks/{item}/place_secondary"
			destroy: str = f"{ns}:custom_blocks/{item}/destroy"
			if placement not in Mem.ctx.data.functions: # Skip if no custom block
				continue

			# If the item is a cable
			if "transfer" in energy:
				write_function(destroy, "# Datapack Energy\nfunction energy:v1/api/break_cable\n", prepend = True)
				write_function(f"{ns}:custom_blocks/{item}/place_secondary", f"""
tag @s add energy.cable
scoreboard players set @s energy.transfer_rate {energy["transfer"]}
function energy:v1/api/init_cable
""")
			else:
				# Else, if if's a machine
				write_function(destroy, "# Datapack Energy\nfunction energy:v1/api/break_machine\n", prepend = True)
				if "usage" in energy or "generation" in energy:
					write_function(f"{ns}:custom_blocks/{item}/place_secondary", f"""
# Energy part
tag @s add energy.{"send" if "generation" in energy else "receive"}
scoreboard players set @s {ns}.energy_rate {energy.get("usage", energy.get("generation", 0))}
scoreboard players set @s energy.max_storage {energy["max_storage"]}
scoreboard players operation @s energy.transfer_rate = @s energy.max_storage
scoreboard players add @s energy.storage 0
scoreboard players add @s energy.change_rate 0
function energy:v1/api/init_machine
""")
				else:
					# Else, it's a battery.
					write_function(f"{ns}:custom_blocks/{item}/place_secondary", f"""
# Energy part
tag @s add {ns}.battery_switcher
tag @s add energy.receive
tag @s add energy.send
data modify storage {ns}:temp energy set from entity @p[tag={ns}.placer] SelectedItem.components."minecraft:custom_data".energy
execute store result score @s energy.max_storage run data get storage {ns}:temp energy.max_storage
execute store result score @s energy.storage run data get storage {ns}:temp energy.storage
scoreboard players operation @s energy.transfer_rate = @s energy.max_storage
function energy:v1/api/init_machine
""")




from .core.dump import dump_command as dump_command
from .core.migrate import migrate_command as migrate_command
from .core.template import template_command as template_command
from .utils import get_project_config as get_project_config
from beet import ProjectConfig as ProjectConfig

def main() -> None: ...




# Imports
from typing import cast

from beet.core.utils import JsonDict, TextComponent


# Page optimizer
def optimize_element(content: TextComponent) -> TextComponent:
	""" Optimize the page content by merging compounds when possible
	Args:
		content (TextComponent): The page content
	Returns:
		TextComponent: The optimized page content
	"""
	# If dict, optimize the values
	if isinstance(content, dict):
		if not any(x in content for x in ["text", "translate", "contents"]):	# If not a text, translate or contents, just return
			return content
		content = content.copy()
		new_component: TextComponent = {}
		for key, value in content.items():
			new_component[key] = optimize_element(value)
		return new_component

	# If not a list, just return
	if not isinstance(content, list):
		return content

	# If list with only one element, return the element
	if len(content) == 1:
		return content[0]

	# For each compound
	new_content: list[TextComponent] = []
	for i, compound in enumerate(content):
		compound = cast(TextComponent, compound)
		if isinstance(compound, list) or i == 0:
			new_content.append(optimize_element(compound))
		else:
			# If the current is a dict with only "text" key, transform it to a string
			if isinstance(compound, dict) and len(compound) == 1 and "text" in compound:
				compound = cast(TextComponent, compound["text"])

			# For checks
			compound_without_text = cast(JsonDict, compound.copy() if isinstance(compound, dict) else compound)
			previous_without_text = cast(JsonDict, new_content[-1].copy() if isinstance(new_content[-1], dict) else new_content[-1])
			if isinstance(compound, dict) and isinstance(new_content[-1], dict):
				compound_without_text.pop("text", None)
				previous_without_text.pop("text", None)

			# If the previous compound is the same as the current one, merge the text
			if str(compound_without_text) == str(previous_without_text):

				# If the previous compound is a text, merge the text
				if isinstance(new_content[-1], str):
					new_content[-1] += str(compound)

				# If the previous compound is a dict, merge the dict
				elif isinstance(new_content[-1], dict):
					new_content[-1]["text"] += cast(JsonDict, compound)["text"]

			# Always add break lines to the previous part (if the text is a string containing only break lines)
			elif isinstance(compound, str) and all(c == "\n" for c in compound):
				if isinstance(new_content[-1], str):
					new_content[-1] += compound
				elif isinstance(new_content[-1], dict):
					new_content[-1]["text"] += compound

			# Always merge two strings
			elif isinstance(compound, str) and isinstance(new_content[-1], str):
				new_content[-1] += compound

			# Otherwise, just add the optimized compound
			else:
				new_content.append(optimize_element(compound))

	# Return
	return new_content

# Remove events recursively
EVENTS: list[str] = ["hover_event", "click_event"]
def remove_events(compound: TextComponent) -> None:
	""" Remove events from a compound recursively
	Args:
		compound (dict): The compound
	"""
	if not isinstance(compound, dict):
		if isinstance(compound, list):
			for element in compound:
				remove_events(element)
		return
	for key in EVENTS:
		if key in compound:
			del compound[key]
	for value in compound.values():
		remove_events(value)

# Function
def optimize_book(book_content: TextComponent) -> TextComponent:
	""" Optimize the book content by associating compounds when possible
	Args:
		book_content (TextComponent): The book content
	Returns:
		TextComponent: The optimized book content
	"""
	if not isinstance(book_content, list):
		book_content = [book_content]

	# For each page, remove events if useless (in an item Lore for example)
	for page in book_content:
		for compound in page:
			if isinstance(compound, list):
				compound_list = cast(list[TextComponent], compound)
			else:
				compound_list = [compound]
			for e in compound_list:
				if isinstance(e, dict):
					e = cast(JsonDict, e)
					for event_type in EVENTS:
						if e.get(event_type):
							remove_events(e[event_type])  # Remove all events below the first event

	# For each page, optimize the array
	return optimize_element(book_content)




""" Quick Sort Algorithm for Minecraft Storage Lists

This algorithm is less efficient than the Selection Sort algorithm for Minecraft due to the cost of macro calls.
It is recommended to use Selection Sort.

Author: Darukshock
"""

# Imports
from beet import Context
from beet.core.utils import JsonDict

from ....core import Mem, write_function
from .constants import MACRO


# Functions
def generate_quick_sort(ctx: Context, sorter: JsonDict) -> None:
	""" Generate quick_sort algorithm functions for sorting storage lists.

	This function creates a set of minecraft functions that implement the quick_sort
	algorithm to sort lists stored in minecraft storage. The algorithm is implemented
	using recursive function calls with macro parameters.

	Args:
		ctx (Context): The beet context for generating functions.
		sorter (dict): Configuration dictionary containing sorting parameters.
	"""
	old_ctx: Context = Mem.ctx
	Mem.ctx = ctx  # Set the current context to the one provided

	# Extract configuration parameters
	functions_location: str = sorter["functions_location"]
	to_sort: str = f"storage {sorter['to_sort']['storage']} {sorter['to_sort']['target']}"
	key: str = sorter["key"]
	scale: int = sorter.get("scale", 1)  # Default scale is 1

	# Main sort entry point function
	write_function(f"{functions_location}/sort", f"""
# Create a scoreboard objective and storage for temporary values
scoreboard objectives add sorter.val dummy
data modify storage sorter:temp inInt set value [0]

# Count the number of elements to sort and prepare the macro
execute store result score high sorter.val run data get {to_sort}
execute store result {MACRO}.high int 1 run scoreboard players remove high sorter.val 1
data modify {MACRO}.low set value 0

# Start the quick_sort process
function {functions_location}/quick_sort with {MACRO}
""")

	# Quicksort recursive function
	write_function(f"{functions_location}/quick_sort", f"""
$execute if predicate {{"condition": "value_check","range": {{"min": -8,"max": $(low)}},"value": $(high)}} run return 0
$execute store result score pi sorter.val run function {functions_location}/partition_main {{low:$(low),high:$(high)}}
data modify storage sorter:temp inInt append value 0
execute store result storage sorter:temp inInt[-1] int 1 store result {MACRO}.high int 1 run scoreboard players remove pi sorter.val 1
function {functions_location}/quick_sort with {MACRO}
execute store result score pi sorter.val run data get storage sorter:temp inInt[-1]
execute store result {MACRO}.low int 1 run scoreboard players add pi sorter.val 2
$data modify {MACRO}.high set value $(high)
function {functions_location}/quick_sort with {MACRO}
""")

	# Partition function - finds pivot position
	write_function(f"{functions_location}/partition_main", f"""
$execute store result score pivot sorter.val run data get {to_sort}[$(high)].{key} {scale}
$scoreboard players set i sorter.val $(low)
scoreboard players remove i sorter.val 1
$execute store result {MACRO}.j int 1 run scoreboard players set j sorter.val $(low)
$function {functions_location}/partition_loop {{high:$(high),low:$(low),j:$(low)}}
execute store result {MACRO}.i int 1 run scoreboard players add i sorter.val 1
$data modify {MACRO}.j set value $(high)
function {functions_location}/partition_swap with {MACRO}
return run scoreboard players get i sorter.val
""")

	# Partition loop - iterates through elements to partition
	write_function(f"{functions_location}/partition_loop", f"""
$execute if score j sorter.val matches $(high).. run return 0
$execute store result score test sorter.val run data get {to_sort}[$(j)].{key} {scale}
execute if score test sorter.val < pivot sorter.val run function {functions_location}/partition_loop_swap with {MACRO}
execute store result {MACRO}.j int 1 run scoreboard players add j sorter.val 1
function {functions_location}/partition_loop with {MACRO}
""")

	# Partition swap - swaps elements during partitioning
	write_function(f"{functions_location}/partition_swap", f"""
$data modify storage sorter:temp element set from {to_sort}[$(i)]
$data modify {to_sort}[$(i)] set from {to_sort}[$(j)]
$data modify {to_sort}[$(j)] set from storage sorter:temp element
""")

	# Partition loop swap - continues partitioning after swap
	write_function(f"{functions_location}/partition_loop_swap", f"""
execute store result {MACRO}.i int 1 run scoreboard players add i sorter.val 1
function {functions_location}/partition_swap with {MACRO}
""")

	# Reset the context to the original one
	Mem.ctx = old_ctx





# Imports
import io
import os
import shutil
import zipfile
from typing import Any

import requests
import stouputils as stp

from ..utils import get_project_config


def migrate_command() -> None:
    """ Handle the 'migrate' command to migrate an existing datapack/resource pack to StewBeet structure.

    This command will:

    1. Check if there's no existing beet configuration
    2. Search for datapack (data/ folder) and resource pack (assets/ folder)
    3. Download and extract the basic template
    4. Migrate existing files to the src/ folder structure

    Ex: `stewbeet migrate`
    """
    # Capture the working directory BEFORE any config check (which might change cwd)
    working_dir = stp.clean_path(os.getcwd())

    # Check if beet configuration already exists
    try:
        get_project_config()
        stp.error("A beet configuration file already exists. Migration is only for projects not using StewBeet or beet yet.")
        return
    except (AssertionError, FileNotFoundError):
        # No config found, continue with migration
        pass

    stp.info("Searching for datapack and resource pack structures...")

    # Search for datapack and resource pack structures
    datapack_info = _find_pack_structure("data", working_dir)
    resource_pack_info = _find_pack_structure("assets", working_dir)

    if not datapack_info and not resource_pack_info:
        stp.error("No datapack or resource pack structure found. Migration requires at least a 'data/' or 'assets/' folder.")
        return

    # Display what was found
    if datapack_info:
        stp.info(f"✓ Found datapack: {stp.relative_path(datapack_info['path'])}")
    if resource_pack_info:
        stp.info(f"✓ Found resource pack: {stp.relative_path(resource_pack_info['path'])}")

    # Ask user confirmation
    stp.warning("This will download the 'basic' template and migrate your existing files to the StewBeet structure.")
    stp.warning("It's recommended to backup your project before proceeding.")
    stp.info("Do you want to continue? (y/n):", end=" ")
    choice = input().strip().lower()
    if choice not in ("y", "yes"):
        stp.info("Migration cancelled.")
        return

    # Download the basic template
    stp.info("Downloading basic template...")
    from importlib.metadata import version
    template_url = f"https://raw.githubusercontent.com/Stoupy51/StewBeet/refs/tags/v{version('stewbeet')}/templates/basic_template.zip"

    try:
        response = requests.get(template_url)
        if response.status_code != 200:
            stp.error(f"Failed to download template. HTTP status code: {response.status_code}")
            return
    except Exception as e:
        stp.error(f"Failed to download template: {e}")
        return

    # Extract the template
    stp.info("Extracting template...")
    try:
        with zipfile.ZipFile(io.BytesIO(response.content)) as zip_file:
            zip_file.extractall(working_dir)
    except Exception as e:
        stp.error(f"Failed to extract template: {e}")
        return

    # Migrate datapack if found
    if datapack_info:
        _migrate_pack(datapack_info, "data", "datapack", working_dir)

    # Migrate resource pack if found
    if resource_pack_info:
        _migrate_pack(resource_pack_info, "assets", "resource pack", working_dir)

    stp.info("✓ Migration completed successfully!")
    stp.info("Next steps:")
    stp.info("  1. Edit beet.yml to configure your project (id, name, author, etc.)")
    stp.info("  2. Run 'stewbeet build' to build your project")


def _find_pack_structure(folder_name: str, search_dir: str | None = None) -> dict[str, Any] | None:
    """ Find a pack structure (data/ or assets/) in the current directory or subdirectories.

    Args:
        folder_name: The folder to search for ('data' or 'assets')
        search_dir: Directory to search in (defaults to current working directory)

    Returns:
        A dict with 'path' (str to the folder), 'parent' (parent directory), and 'has_pack_mcmeta' (bool)
        Returns None if not found or if pack.mcmeta doesn't exist alongside
    """
    current_dir = search_dir if search_dir else os.getcwd()

    # Check if folder exists at root level
    folder_path = f"{current_dir}/{folder_name}"
    if os.path.isdir(folder_path):
        pack_mcmeta_path = f"{current_dir}/pack.mcmeta"
        has_pack_mcmeta = os.path.exists(pack_mcmeta_path)
        if has_pack_mcmeta:
            return {
                "path": folder_path,
                "parent": current_dir,
                "has_pack_mcmeta": has_pack_mcmeta
            }

    # Check in subdirectories (up to two levels deep)
    for subdir_name in os.listdir(current_dir):
        subdir = f"{current_dir}/{subdir_name}"
        if os.path.isdir(subdir):
            # Check one level deep
            pack_folder = f"{subdir}/{folder_name}"
            if os.path.isdir(pack_folder):
                pack_mcmeta_path = f"{subdir}/pack.mcmeta"
                has_pack_mcmeta = os.path.exists(pack_mcmeta_path)
                if has_pack_mcmeta:
                    return {
                        "path": pack_folder,
                        "parent": subdir,
                        "has_pack_mcmeta": has_pack_mcmeta
                    }

            # Check two levels deep
            for subsubdir_name in os.listdir(subdir):
                subsubdir = f"{subdir}/{subsubdir_name}"
                if os.path.isdir(subsubdir):
                    pack_folder = f"{subsubdir}/{folder_name}"
                    if os.path.isdir(pack_folder):
                        pack_mcmeta_path = f"{subsubdir}/pack.mcmeta"
                        has_pack_mcmeta = os.path.exists(pack_mcmeta_path)
                        if has_pack_mcmeta:
                            return {
                                "path": pack_folder,
                                "parent": subsubdir,
                                "has_pack_mcmeta": has_pack_mcmeta
                            }

    return None


def _migrate_pack(pack_info: dict[str, str | bool], folder_name: str, pack_type: str, working_dir: str) -> None:
    """ Migrate a pack (datapack or resource pack) to the StewBeet structure.

    Args:
        pack_info: Dict containing pack information from _find_pack_structure
        folder_name: 'data' or 'assets'
        pack_type: 'datapack' or 'resource pack' (for logging)
        working_dir: The working directory where migration is happening
    """
    stp.info(f"Migrating {pack_type}...")

    # Remove the template's src folder for this pack type if it exists
    template_src_folder = f"{working_dir}/src/{folder_name}"
    if os.path.exists(template_src_folder):
        stp.debug(f"  Removing template's {stp.relative_path(template_src_folder)}...")
        shutil.rmtree(template_src_folder)

    # Create src directory if it doesn't exist
    src_dir = f"{working_dir}/src"
    os.makedirs(src_dir, exist_ok=True)

    # Move the pack folder to src/
    source_path: str = pack_info["path"]  # type: ignore
    dest_path = f"{src_dir}/{folder_name}"

    stp.info(f"  Moving {stp.relative_path(source_path)} to {stp.relative_path(dest_path)}...")
    try:
        shutil.move(source_path, dest_path)
    except Exception as e:
        stp.error(f"Failed to move {folder_name}: {e}")
        return

    # If pack.mcmeta exists in the parent directory and we're in a subdirectory, move it too
    parent_path: str = pack_info["parent"]  # type: ignore
    has_pack_mcmeta: bool = pack_info["has_pack_mcmeta"]  # type: ignore
    if has_pack_mcmeta and parent_path != working_dir:
        pack_mcmeta_src = f"{parent_path}/pack.mcmeta"
        pack_mcmeta_dest = f"{src_dir}/pack.mcmeta"
        if os.path.exists(pack_mcmeta_src) and not os.path.exists(pack_mcmeta_dest):
            stp.debug(f"  Moving pack.mcmeta to {stp.relative_path(pack_mcmeta_dest)}...")
            shutil.move(pack_mcmeta_src, pack_mcmeta_dest)

        # Also move pack.png if it exists (goes to assets/ folder at root)
        pack_png_src = f"{parent_path}/pack.png"
        assets_dir = f"{working_dir}/assets"
        os.makedirs(assets_dir, exist_ok=True)
        pack_png_dest = f"{assets_dir}/pack.png"
        if os.path.exists(pack_png_src) and not os.path.exists(pack_png_dest):
            stp.debug(f"  Moving pack.png to {stp.relative_path(pack_png_dest)}...")
            shutil.move(pack_png_src, pack_png_dest)

    # Clean up empty parent directory if it was a subdirectory
    if parent_path != working_dir:
        try:
            # Check if directory is empty
            if not os.listdir(parent_path):
                stp.debug(f"  Removing empty directory {stp.relative_path(parent_path)}...")
                os.rmdir(parent_path)
        except Exception:
            pass  # Ignore errors when cleaning up

    stp.info(f"  ✓ {pack_type.capitalize()} migrated successfully")




# pyright: reportGeneralTypeIssues=false
# ruff: noqa: RUF012
# Imports
import json
import os
from enum import Enum

import stouputils as stp
from beet import ItemModel, Model
from beet.core.utils import JsonDict

from ...core import Mem, texture_mcmeta


# GUI Translation for
class GuiTranslation(Enum):
	brewing_stand: list[int] = [ 0, 38, -76 ]
	""" Assume you are placing GUI item model in the __fuel__ slot of a brewing stand. """
	furnace_bottom: list[int] = [ 0, 75, -76 ]
	""" Assume you are placing GUI item model in the __fuel__ slot of a furnace. """
	furnace_top: list[int] = [ 0, 39, -76 ]
	""" Assume you are placing GUI item model in the __ingredient__ slot of a furnace. """
	barrel_bottom_right: list[int] = [ -78, 75, -76 ]
	""" Assume you are placing GUI item model in the __27th__ slot of a barrel. """

# Setup GUI in resource packs
def setup_gui_in_resource_packs(gui_translations: dict[str, GuiTranslation]) -> dict[str, str]:
	""" Setup GUI item models in resource packs by creating item models and textures for each GUI.

	Args:
		gui_translations	(dict[str, GuiTranslation]): A dictionary mapping GUI names to their translation offsets.
			(e.g. {'electric_brewing_stand': GuiTranslation.brewing_stand, 'electric_furnace': GuiTranslation.furnace_bottom, ...})
	Returns:
		dict[str, str]: A dictionary mapping GUI filenames to their model paths.
			(e.g. {'gui/electric_brewing_stand.png': 'namespace:gui/electric_brewing_stand'})
	"""
	namespace: str = Mem.ctx.project_id
	textures_folder: str = Mem.ctx.meta.get("stewbeet", {}).get("textures_folder", "")

	# List gui asset filenames and map with 'gui/{filename}' keys
	filenames: list[str] = os.listdir(f"{textures_folder}/gui")
	gui_models: dict[str, str] = {f"gui/{x}": f"{namespace}:gui/{x.replace('.png', '')}" for x in filenames if x.endswith(".png")}

	# Write custom models
	base: JsonDict = {
		"textures" : {},
		"elements": [
			{
				"from": [ -16, -16, 15.9375 ],
				"to": [ 32, 32, 16 ],
				"faces": {
					"north": { "uv": [ 11, 16, 0, 5 ], "rotation": 180, "texture": "#layer0" },
					"south": { "uv": [ 0, 5, 11, 16 ], "texture": "#layer0" }
				}
			}
		],
		"display": {
			"firstperson_lefthand": {"rotation": [ 0, 0, 0 ],"translation": [ 0, 0, 0 ],"scale": [ 0, 0, 0 ]},
			"gui": {"rotation": [ 0, 0, 0 ],"scale": [ 3.66, 3.66, 3.66 ]},
			"ground": {"rotation": [ 0, 0, 0 ],"translation": [ 0, 0, 0 ],"scale": [ 0, 0, 0 ]}
		},
		"gui_light":"front"
	}
	for gui, model in gui_models.items():
		content: JsonDict = json.loads(json.dumps(base))  # Deep copy the base model
		content["textures"]["layer0"] = content["textures"]["particle"] = model.replace(":", ":item/")

		# If any translation is provided, add it to the model
		for item, translation in gui_translations.items():
			if item in gui:
				content["display"]["gui"]["translation"] = translation.value
				break

		# Else, remove the gui display and set parent to item/generated
		else:
			content["display"].pop("gui")
			content.pop("elements")
			content["parent"] = "item/generated"

		# Write the model in models/item/
		model_name = gui.replace('.png', '')
		Mem.ctx.assets[namespace].models[f"item/{model_name}"] = Model(
			stp.json_dump(content, max_level=3)
		)

		# Copy the textures
		Mem.ctx.assets[namespace].textures[f"item/{model_name}"] = texture_mcmeta(f"{textures_folder}/{gui}")

		# Write the file in items/
		Mem.ctx.assets[namespace].item_models[model_name] = ItemModel(
			stp.json_dump({
				"model": {
					"type": "minecraft:model",
					"model": f"{namespace}:item/{model_name}"
				},
				"oversized_in_gui": True
			}, max_level=3)
		)

	return gui_models





# Imports
from beet.core.utils import TextComponent

from ..__memory__ import Mem
from ..cls.item import Item


def format_energy_number(number: int) -> str:
	""" Formats a number into a string with appropriate unit suffix (k, M, G, T).

	Args:
		number (int): The number to format
	Returns:
		str: Formatted string with unit suffix, ex: 12000 -> "12 M", 1200 -> "1200 k"
	"""
	if number < 9999:
		return f"{number} k"
	elif number < 9999999:
		return f"{number/1000:.0f} M"
	elif number < 9999999999:
		return f"{number/1000000:.0f} G"
	else:
		return f"{number/1000000000:.0f} T"

def create_energy_lore(energy_data: dict[str, int]) -> TextComponent:
	""" Creates lore entries for energy-related blocks based on their energy data.

	Args:
		energy_data (dict): Dictionary containing energy configuration values
	Returns:
		TextComponent: Formatted text component with energy lore
	"""
	# Determine if this is a battery (no usage/generation) or a cable (only transfer)
	is_battery: bool = "max_storage" in energy_data and "usage" not in energy_data and "generation" not in energy_data
	is_cable: bool = "transfer" in energy_data and "max_storage" not in energy_data

	# Create the lore config
	lore_config: dict[str, tuple[str, str]] = {
		"generation": ("Energy Generation", "W"),
		"usage": ("Power Usage", "W"),
		"max_storage": ("Energy Storage" if is_battery else "Energy Buffer", "J"),
		"transfer": ("Transfer Speed" if is_cable else "Energy Transfer", "W")
	}

	# Create the lore entries
	lore: TextComponent = []
	for key, (label, unit) in lore_config.items():
		if key in energy_data:
			lore.append({
				"text": f"[{label}: {format_energy_number(energy_data[key])}{unit}]",
				"italic": False,
				"color": "gray"
			})

	return lore

def add_energy_lore_to_definitions():
	""" Adds energy-related lore to definitions based on their custom data. """
	for data in Mem.definitions.values():
		if isinstance(data, Item):
			data = data.components
		if "energy" in data.get("custom_data", {}):
			data["lore"] = data.get("lore", []) + create_energy_lore(data["custom_data"]["energy"])




import stouputils as stp
from beet.core.utils import JsonDict as JsonDict

def validate_config(pmc_config: dict[str, str]) -> str:
    """ Validate PlanetMinecraft configuration

\tArgs:
\t\tpmc_config (dict[str, str]): Configuration for the PlanetMinecraft project
\tReturns:
\t\tstr: Project url on PlanetMinecraft
\t"""
def convert_markdown_to_bbcode(markdown: str, verbose: bool = True) -> str:
    """ Convert markdown to bbcode for PlanetMinecraft

\tArgs:
\t\tmarkdown (str): Markdown text
\t\tverbose (bool): If True, print the conversion comparison

\tReturns:
\t\tstr: BBcode text

\tExamples:
\t\t>>> markdown_text = '''
\t\t... ## Changelog
\t\t...
\t\t... ### Build System
\t\t... - 🚀 Bump version to v1.2.3 ([2111fd2](https://github.com/Stoupy51/LifeSteal/commit/2111fd2f390b80a3aab77a4e7bcbb24b93845e5a))
\t\t...
\t\t... ### Features
\t\t... - ✨ Added new configuration for dropping heart (non pvp) ([cde8749](https://github.com/Stoupy51/LifeSteal/commit/cde8749aa9e447302481f50b9887a0b3a846c7fe))
\t\t...
\t\t... **Full Changelog**: https://github.com/Stoupy51/LifeSteal/compare/v1.2.2...v1.2.3
\t\t... '''
\t\t>>> bbcode = convert_markdown_to_bbcode(markdown_text, verbose=False)
\t\t>>> print(bbcode.strip())
\t\t[h2]Changelog[/h2][h4]Build System[/h4][list]
\t\t[*]🚀 Bump version to v1.2.3 ([url=https://github.com/Stoupy51/LifeSteal/commit/2111fd2f390b80a3aab77a4e7bcbb24b93845e5a]2111fd2[/url])[/*]
\t\t[/list][h4]Features[/h4][list]
\t\t[*]✨ Added new configuration for dropping heart (non pvp) ([url=https://github.com/Stoupy51/LifeSteal/commit/cde8749aa9e447302481f50b9887a0b3a846c7fe]cde8749[/url])[/*]
\t\t[/list]
\t\t[b]Full Changelog[/b]: [url]https://github.com/Stoupy51/LifeSteal/compare/v1.2.2...v1.2.3[/url]
\t"""
def upload_version(project_url: str, changelog: str) -> None:
    """ Upload new version by opening the project url with the browser

\tArgs:
\t\tproject_url\t\t(str):\tUrl of the project on PlanetMinecraft to open
\t\tchangelog\t\t(str):\tChangelog text
\t"""
@stp.handle_error
def upload_to_pmc(pmc_config: JsonDict, changelog: str = '') -> None:
    """ Upload the project to PlanetMinecraft using the configuration

\tDisclaimer:
\t\tThere is no API for PlanetMinecraft, so everything is done manually.
\tArgs:
\t\tpmc_config\t\t(dict):\t\tConfiguration for the PlanetMinecraft project
\t\tchangelog\t\t(str):\t\tChangelog text for the release
\t"""




# Imports
import os
import sys
import zipfile

import stouputils as stp

from ..utils import get_project_config


def dump_command() -> None:
    """ Handle the 'dump' command to create a zip archive of the project.
    Excludes build outputs, cache directories, and other temporary files.

    Ex: `stewbeet dump [output_name.zip]`
    """
    # Get the project configuration
    cfg = get_project_config()

    # Determine output zip filename
    output_name: str = sys.argv[2] if len(sys.argv) >= 3 else "project_dump.zip"
    if not output_name.endswith(".zip"):
        output_name += ".zip"

    # Get paths to exclude from the archive
    exclude_paths: set[str] = {
        ".beet_cache",
        "__pycache__",
        ".git",
        ".vscode",
        ".idea",
        "*.pyc",
        ".DS_Store",
        "Thumbs.db",
    }

    # Add patterns from .gitignore if it exists
    if os.path.exists(".gitignore"):
        with stp.super_open(".gitignore", "r") as f:
            for line in f.readlines():
                line: str = line.strip()
                # Skip empty lines and comments
                if line and not line.startswith("#"):
                    exclude_paths.add(line)

    # Add output directory from config
    if cfg.output:
        exclude_paths.add(stp.relative_path(str(cfg.output)))

    # Add manual cache path from config
    manual_config = cfg.meta.get("stewbeet", {}).get("manual", {})
    cache_path: str = manual_config.get("cache_path", "")
    if cache_path:
        exclude_paths.add(stp.relative_path(cache_path))

    # Add definitions_debug if specified
    definitions_debug: str = cfg.meta.get("stewbeet", {}).get("definitions_debug", "")
    if definitions_debug:
        exclude_paths.add(stp.relative_path(definitions_debug))

    # Add files from ignore list in config
    ignore_patterns: list[str] = cfg.ignore or []
    exclude_paths.update(ignore_patterns)

    stp.debug(f"Creating project archive: '{output_name}'")
    stp.debug(f"Excluding: {', '.join(sorted(exclude_paths))}")

    # Create the zip file
    files_added = 0

    try:
        with zipfile.ZipFile(output_name, "w", zipfile.ZIP_DEFLATED) as zip_file:
            # Walk through all files in the current directory
            for root, dirs, files in os.walk("."):
                # Get relative path
                root = stp.relative_path(root)

                # Check if this directory should be excluded
                should_exclude = False
                for exclude in exclude_paths:
                    # Check if the current path matches any exclude pattern
                    if exclude in root.split("/"):
                        should_exclude = True
                        break
                    # Check if the path starts with the exclude pattern
                    if root.startswith(exclude.rstrip("*")):
                        should_exclude = True
                        break

                if should_exclude:
                    # Remove excluded directories from dirs list to prevent os.walk from traversing them
                    dirs[:] = []
                    continue

                # Filter out excluded directories from dirs list
                dirs[:] = [
                    d for d in dirs
                    if not any(
                        d == excl or d.startswith(excl.rstrip("*"))
                        for excl in exclude_paths
                    )
                ]

                # Add files from this directory
                for file in files:
                    file_path = f"{root}/{file}" if root != "." else file

                    # Check if this specific file should be excluded
                    should_exclude_file = False
                    for exclude in exclude_paths:
                        if exclude.startswith("*") and file_path.endswith(exclude[1:]):
                            should_exclude_file = True
                            break
                        if file_path == exclude or file_path.startswith(exclude):
                            should_exclude_file = True
                            break

                    # Don't include the output zip file itself
                    if file_path == output_name:
                        should_exclude_file = True

                    if not should_exclude_file:
                        zip_file.write(file_path, file_path)
                        files_added += 1

        stp.info(f"✓ Successfully created archive with {files_added} files: '{output_name}'")

    except Exception as e:
        stp.error(f"Failed to create archive: {e}")
        # Remove partially created zip file
        if os.path.exists(output_name):
            os.remove(output_name)
        raise





# Imports
from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, fields
from typing import Any, Self

import stouputils as stp
from beet.core.utils import JsonDict

from ..constants import NOT_COMPONENTS


# Class for mapping behavior
@dataclass
class StMapping(Mapping[str, Any]):
    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)
    def __setitem__(self, key: str, value: Any) -> None:
        return setattr(self, key, value)
    def __contains__(self, key: Any) -> bool:
        return self.get(key) is not None

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self[key]
        except (KeyError, AttributeError):
            return default

    def setdefault(self, key: str, default: Any = None) -> Any:
        """ Set a default value if key doesn't exist, like dict.setdefault(). """
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def to_dict(self) -> JsonDict:
        """ Convert the object to a dictionary for JSON serialization """
        from dataclasses import asdict, is_dataclass

        def _convert_value(value: Any) -> Any:
            """ Recursively convert a value to a JSON-serializable form """
            if value is None:
                return None
            elif hasattr(value, 'to_dict'):
                return value.to_dict()
            elif is_dataclass(value) and not isinstance(value, type):
                return asdict(value)
            elif isinstance(value, list):
                return [_convert_value(item) for item in value] # pyright: ignore[reportUnknownVariableType]
            elif isinstance(value, dict):
                return {k: _convert_value(v) for k, v in value.items()} # pyright: ignore[reportUnknownVariableType]
            else:
                return value

        result: JsonDict = {}
        for field_info in fields(self):
            value = getattr(self, field_info.name)
            if value is not None:
                result[field_info.name] = _convert_value(value)
        return result

    @classmethod
    def from_dict(cls, data: JsonDict | StMapping, item_id: str) -> Self:
        """ Create an object based on items """
        if isinstance(data, StMapping):
            return data # type: ignore

        # Make a copy to avoid modifying the original
        # Rename some fields from StewBeet v2.x to v3.x
        rename_dict: dict[str, str] = {
            "id": "base_item",
            "category": "manual_category",
            "result_of_crafting": "recipes",
            "used_for_crafting": "recipes",
            "wiki_components": "wiki_buttons",
        }
        data_dict: JsonDict = dict(data)
        for old, new in rename_dict.items():
            if old in data_dict and new not in data_dict:
                data_dict[new] = data_dict.pop(old)
            elif old in data_dict and new in data_dict:
                if data_dict[new] == data_dict[old]:
                    data_dict.pop(old)
                elif isinstance(data_dict[new], list) and isinstance(data_dict[old], list):
                    data_dict[new] = stp.unique_list([*data_dict[new], *data_dict.pop(old)])
                else:
                    pass
        data_dict["id"] = item_id

        # Get valid field names for this class
        valid_fields: set[str] = {f.name for f in fields(cls)}

        # Separate known fields from unknown fields
        known_kwargs: JsonDict = {}
        unknown_kwargs: JsonDict = {}
        for key, value in data_dict.items():
            if key in valid_fields:
                known_kwargs[key] = value
            else:
                unknown_kwargs[key] = value

        # If there are unknown fields and the class has a 'components' field, add them there
        if unknown_kwargs and "components" in valid_fields:
            # Merge with existing components if any
            existing_components = known_kwargs.get('components', {})
            if isinstance(existing_components, dict):
                known_kwargs["components"] = {**existing_components, **unknown_kwargs}
            else:
                known_kwargs["components"] = unknown_kwargs
        elif unknown_kwargs:
            # If no components field exists, raise an error
            raise TypeError(f"{cls.__name__}() got unexpected keyword arguments: {', '.join(unknown_kwargs.keys())}")

        # Remove unexpected components keys (from StewBeet)
        for key in NOT_COMPONENTS:
            if "components" in known_kwargs and key in known_kwargs["components"]:
                del known_kwargs["components"][key]

        # Add empty vanilla_block if needed
        if "vanilla_block" in valid_fields:
            if "vanilla_block" not in known_kwargs:
                known_kwargs["vanilla_block"] = ""

        # Create the instance
        return cls(**known_kwargs)

    @classmethod
    def from_id(cls, item_id: str, strict: bool = True) -> Self:
        """ Create an object based of definitions. If ':' is in item_id, it's in external_definitions

        Args:
            item_id	(str):		The item ID to create the object from.
            strict	(bool):		Whether to raise an error if the item is not found.
        """
        from ..__memory__ import Mem
        if strict:
            if ":" not in item_id:
                return cls.from_dict(Mem.definitions[item_id], item_id)
            else:
                return cls.from_dict(Mem.external_definitions[item_id], item_id)
        else:
            if ":" not in item_id:
                return cls.from_dict(Mem.definitions.get(item_id, {}), item_id)
            else:
                return cls.from_dict(Mem.external_definitions.get(item_id, {}), item_id)

    @classmethod
    def clone(cls, other: Self) -> Self:
        """ Create a clone of another instance """
        return cls(**other.to_dict())

    def copy(self) -> JsonDict:
        """ Return a shallow copy as a dictionary. """
        return self.to_dict()

    # Mapping methods (__len__ and __iter__)
    def __len__(self) -> int:
        return len(self.to_dict())
    def __iter__(self):
        return iter(self.to_dict())





# Imports
import stouputils as stp
from beet import Context

from ....core.__memory__ import Mem
from ....core.utils.io import write_tick_file, write_versioned_function


# Main entry point
@stp.measure_time(message="Execution time of 'stewbeet.plugins.finalyze.basic_datapack_structure'")
def beet_default(ctx: Context) -> None:
	""" Main entry point for the basic datapack structure plugin.
	This plugin sets up basic timing structures for a Minecraft datapack with tick functions
	for different intervals (tick_2, second, second_5, minute).

	Args:
		ctx (Context): The beet context.
	"""
	if Mem.ctx is None: # pyright: ignore[reportUnnecessaryComparison]
		Mem.ctx = ctx

	# Get namespace and version
	assert ctx.project_id, "Project ID is not set. Please set it in the project configuration."
	ns: str = ctx.project_id
	version: str = ctx.project_version

	# Define function paths
	tick_2_path: str = f"{ns}:v{version}/tick_2"
	second_path: str = f"{ns}:v{version}/second"
	second_5_path: str = f"{ns}:v{version}/second_5"
	minute_path: str = f"{ns}:v{version}/minute"

	# Check if functions exist in the data pack
	tick_2_exists: bool = tick_2_path in ctx.data.functions
	second_exists: bool = second_path in ctx.data.functions
	second_5_exists: bool = second_5_path in ctx.data.functions
	minute_exists: bool = minute_path in ctx.data.functions

	# Prepend timer reset to existing functions
	if tick_2_exists:
		write_versioned_function("tick_2",
f"""# Reset timer
scoreboard players set #tick_2 {ns}.data 1
""", prepend=True)

	if second_exists:
		write_versioned_function("second",
f"""# Reset timer
scoreboard players set #second {ns}.data 0
""", prepend=True)

	if second_5_exists:
		write_versioned_function("second_5",
f"""# Reset timer
scoreboard players set #second_5 {ns}.data -10
""", prepend=True)

	if minute_exists:
		write_versioned_function("minute",
f"""# Reset timer
scoreboard players set #minute {ns}.data 1
""", prepend=True)
	# Generate tick structure if any timer functions exist
	if tick_2_exists or second_exists or second_5_exists or minute_exists:
		content: str = "# Timers\n"

		# Add scoreboard increments for existing timers
		if tick_2_exists:
			content += f"scoreboard players add #tick_2 {ns}.data 1\n"
		if second_exists:
			content += f"scoreboard players add #second {ns}.data 1\n"
		if second_5_exists:
			content += f"scoreboard players add #second_5 {ns}.data 1\n"
		if minute_exists:
			content += f"scoreboard players add #minute {ns}.data 1\n"

		# Add function calls when timers reach their thresholds
		# tick_2 and second_5 are "offsync" for better load distribution
		if tick_2_exists:
			content += f"execute if score #tick_2 {ns}.data matches 3.. run function {ns}:v{version}/tick_2\n"
		if second_exists:
			content += f"execute if score #second {ns}.data matches 20.. run function {ns}:v{version}/second\n"
		if second_5_exists:
			content += f"execute if score #second_5 {ns}.data matches 90.. run function {ns}:v{version}/second_5\n"
		if minute_exists:
			content += f"execute if score #minute {ns}.data matches 1200.. run function {ns}:v{version}/minute\n"

		# Write the timer logic to the tick file
		if content:
			write_tick_file(content, prepend=True)





# Imports
from beet import Context  # pyright: ignore[reportUnusedImport] # noqa: F401

from .balancing import *
from .batteries import *
from .cables import *
from .energy_lib_calls import *
from .gui import *
from .wrench import *





# Imports
import os
from pathlib import Path
from typing import Any, cast

import stouputils as stp
from beet import Context, Texture
from beet.core.utils import JsonDict


# Main entry point
@stp.measure_time(message="Execution time of 'stewbeet.plugins.finalyze.check_unused_textures'")
def beet_default(ctx: Context) -> None:
	""" Main entry point for the check unused textures plugin.
	This plugin checks for unused textures in the resource pack by analyzing all JSON files
	and comparing texture references with the available texture files.

	Args:
		ctx (Context): The beet context.
	"""
	# Assertions
	stewbeet: JsonDict = ctx.meta.get("stewbeet", {})
	textures_folder: str = stp.clean_path(stewbeet.get("textures_folder", ""))
	assert textures_folder, "meta.stewbeet.textures_folder is not set. Please set it in the project configuration."

	# 1) Build a dict of all textures file paths relative to the textures folder:
	# Ex: {'some_folder/dirt.png', 'stone.png', ...}
	textures: set[str] = {stp.relative_path(str(p), textures_folder) for p in Path(textures_folder).rglob("*.png")}

	# 2) For each texture, check if any of the ctx.assets.textures matches the texture filename.
	unused_paths: set[str] = set()
	for path in textures:
		# Get just the filename without extension for comparison
		filename_no_ext: str = os.path.splitext(os.path.basename(path))[0]
		no_extension_path: str = os.path.splitext(path)[0]
		if not any(
			(str(texture.source_path).endswith(no_extension_path) or filename_no_ext in str(texture.source_path)) if isinstance(texture, Texture)
			else (
				(texture.endswith(no_extension_path) or filename_no_ext in texture) if isinstance(texture, str)
				else False
			)
			for texture in cast(list[Any], ctx.assets.textures)
		):
			unused_paths.add(path)

	# 3) If anything is unused, warn about it:
	if unused_paths:
		warning_lines: list[str] = [
			f"- '{textures_folder}/{path}'"
			for path in sorted(unused_paths)
		]
		warning_msg: str = (
			"Some textures are not used in the resource pack:\n"
			+ "\n".join(warning_lines)
		)
		stp.warning(warning_msg)





# Imports
from pathlib import Path

import stouputils as stp
from beet import Atlas, Context, ResourcePack
from beet.core.utils import JsonDict

from ....core.__memory__ import Mem
from ....core.cls.item import Item
from .object import AutoModel, to_atlas


# Utility function to add textures to atlas
def add_to_atlas(textures: set[str] = set()) -> None:  # noqa: B006
	""" Add textures to the specified atlas.

	Args:
		textures	(set[str]):	The set of texture paths to add. Defaults to an empty set.
	"""
	if not textures:
		return

	# Get mcmeta and setup overlays if needed
	rp: ResourcePack = Mem.ctx.assets

	# Define overlay configurations
	overlay_configs: list[JsonDict] = [
		{
			"directory": "before_format_73",
			"formats": [0, 72],
			"min_format": 0,
			"max_format": 72,
			"atlas_name": "blocks"
		},
		{
			"directory": "format_73_plus",
			"formats": [73, 1000],
			"min_format": 73,
			"max_format": 1000,
			"atlas_name": "items"
		}
	]

	# Process each overlay configuration
	for config in overlay_configs:
		directory: str = config["directory"]

		# Check if this overlay already exists
		rp.overlays[directory] = ResourcePack(
			name=directory,
			supported_formats=config["formats"],
			min_format=config["min_format"],
			max_format=config["max_format"]
		)

		# Add textures to the appropriate atlas in the overlay
		overlay = Mem.ctx.assets.overlays[directory]
		atlas_object: Atlas = overlay["minecraft"].atlases.setdefault(config["atlas_name"])
		data: JsonDict = atlas_object.data
		sources: list[JsonDict] = data.get("sources", [])

		for texture in textures:
			sources.append({"type": "minecraft:single", "resource": texture, "sprite": to_atlas(texture)})

		sources = stp.unique_list(sorted(sources, key=lambda x: x["resource"]))
		atlas_object.data["sources"] = sources
		atlas_object.encoder = stp.json_dump


# Main entry point
@stp.measure_time(message="Execution time of 'stewbeet.plugins.resource_pack.item_models'")
def beet_default(ctx: Context):
	""" Main entry point for the item models plugin.

	Args:
		ctx (Context): The beet context.
	"""
	## Assertions
	# Stewbeet Initialized
	if Mem.ctx is None: # pyright: ignore[reportUnnecessaryComparison]
		Mem.ctx = ctx

	# Textures folder
	textures_folder: str = stp.relative_path(Mem.ctx.meta.get("stewbeet", {}).get("textures_folder", ""))
	assert textures_folder != "", "Textures folder path not found in 'ctx.meta.stewbeet.textures_folder'. Please set a directory path in project configuration."

	# Textures
	textures: dict[str, str] = {
		stp.clean_path(str(p)).split("/")[-1]: stp.relative_path(str(p))
		for p in Path(textures_folder).rglob("*.png")
	}

	# Initialize rendered_item_models set in ctx.meta
	Mem.ctx.meta["stewbeet"]["rendered_item_models"] = set()

	# Get all item models from definitions
	item_models: dict[str, AutoModel] = {}
	for item_name in Mem.definitions.keys():
		obj = Item.from_id(item_name)

		# Skip items without models or already rendered
		item_model: str = obj.components.get("item_model", "")
		if not item_model or item_model in Mem.ctx.meta["stewbeet"]["rendered_item_models"]:
			continue

		# Skip items not in our namespace
		if not item_model.startswith(Mem.ctx.project_id):
			continue

		# Create an MyItemModel object from the definitions entry
		item_models[item_name] = AutoModel.from_definitions(obj, textures)

	# Process each item model
	used_minecraft_textures: set[str] = set()
	for model in item_models.values():
		used_minecraft_textures.update(model.process())

	# If any of the minecraft textures used are not in the items atlas, add them
	not_in_atlas: set[str] = {texture for texture in used_minecraft_textures if not texture.startswith("minecraft:item/")}
	add_to_atlas(textures=not_in_atlas)





# Imports
import os

import stouputils as stp
from PIL import Image

from ...core.__memory__ import Mem
from .image_utils import careful_resize
from .shared_import import SharedMemory


# Functions
def calculate_optimal_grid(item_count: int) -> tuple[int, int]:
	""" Calculate optimal grid dimensions for a given number of items to achieve closest to 16:9 ratio

	Args:
        item_count (int): The total number of items to arrange in the grid.
    Returns:
        tuple[int, int]: The optimal number of rows and columns for the grid.
	"""
	if item_count == 0:
		return 0, 0

	best_ratio_diff: float = float('inf')
	best_rows, best_cols = 1, item_count
	target_ratio: float = 16/9

	# Try different configurations
	for rows in range(1, item_count + 1):
		cols: int = (item_count + rows - 1) // rows  # Ceiling division
		if rows * cols >= item_count:
			ratio: float = cols / rows
			ratio_diff: float = abs(ratio - target_ratio)
			if ratio_diff < best_ratio_diff:
				best_ratio_diff = ratio_diff
				best_rows, best_cols = rows, cols

	return best_rows, best_cols

def generate_showcase_images(showcase_mode: int, categories: dict[str, list[str]], simple_case: Image.Image):
	""" Generate showcase images based on the showcase_mode parameter

    Args:
        showcase_mode   (int): Mode for generating showcase images:
            1 - Showcase items in the manual
            2 - Showcase all items, even those not in the manual
            3 - Showcase both manual items and all items
        categories      (dict[str, list]): Dictionary of categories with items.
        simple_case     (Image.Image): Image of a simple case to use as background for items.
	"""
	# Get items for manual (mode 1 or 3)
	if showcase_mode in [1, 3]:
		manual_items: list[str] = []
		for items in categories.values():
			manual_items.extend(items)
		if manual_items:
			stp.run_in_subprocess(
				create_showcase_image, manual_items, "all_manual_items.png", simple_case,
				str(Mem.ctx.output_directory), Mem.ctx.project_id, SharedMemory.cache_path,
				no_join=True
			)

	# Get all items (mode 2 or 3)
	if showcase_mode in [2, 3]:
		all_items: list[str] = list(Mem.definitions.keys())
		if all_items:
			stp.run_in_subprocess(
				create_showcase_image, all_items, "all_items.png", simple_case,
				str(Mem.ctx.output_directory), Mem.ctx.project_id, SharedMemory.cache_path,
				no_join=True
			)

def create_showcase_image(
	items: list[str], filename: str, simple_case: Image.Image, output_dir: str, project_id: str, cache_path: str
) -> None:
	""" Create a showcase image with items arranged in optimal grid for 16:9 ratio

    Args:
        items           (list[str]): List of item IDs to include in the showcase.
        filename        (str): Name of the output image file.
        simple_case     (Image.Image): Image of a simple case to use as background for items.
		output_dir      (str): Directory to save the output image.
		project_id      (str): Project identifier for path construction.
		cache_path      (str): Path to the cache directory for item textures.
	"""
	if not items:
		return

	rows, cols = calculate_optimal_grid(len(items))

	# Big case size (keeps textures sharp)
	case_size: int = 512

	# Pre-resize the case once, keep mode consistent with final image (RGBA)
	resized_case: Image.Image = simple_case.convert("RGBA").resize((case_size, case_size), Image.Resampling.NEAREST)

	# Canvas size
	img_width = cols * case_size
	img_height = rows * case_size

	# Create canvas (transparent background)
	showcase_image = Image.new("RGBA", (img_width, img_height), (0, 0, 0, 0))

	# Tile the background: paste resized_case for every grid cell once
	# Doing this separately avoids repeated case pastes mixed with item processing.
	for r in range(rows):
		y = r * case_size
		for c in range(cols):
			x = c * case_size
			# use the C-level paste; cheap and fast
			showcase_image.paste(resized_case, (x, y))

	# Cache for textures to avoid reopening/resizing the same file multiple times
	# Keyed by texture_path; stores PIL Image (already resized to target_size, RGBA)
	texture_cache: dict[str, Image.Image] = {}

	# Pre-calc item target size (89% of case)
	target_size = int(case_size * 0.890625)

	# Process and paste items
	for i, item in enumerate(items):
		row = i // cols
		col = i % cols

		x = col * case_size
		y = row * case_size

		texture_path = f"{cache_path}/items/{project_id}/{item}.png"

		# Load (and cache) resized texture
		resized_item = texture_cache.get(texture_path)
		if resized_item is None:
			try:
				# Attempt to open and convert to RGBA immediately.
				# Using "with" ensures file handles are closed promptly.
				with Image.open(texture_path) as img:
					img_rgba = img.convert("RGBA")
					# careful_resize is presumably tuned for preserving pixel art; use it.
					resized_item = careful_resize(img_rgba, target_size)
			except (FileNotFoundError, OSError):
				stp.warning(f"Missing texture at '{texture_path}', using empty texture for showcase")
				# Create an appropriately sized transparent image instead of 1x1 to avoid repeated resizing
				resized_item = Image.new("RGBA", (target_size, target_size), (0, 0, 0, 0))

			# Store in cache
			texture_cache[texture_path] = resized_item

		# Center item in the case cell
		item_x = x + (case_size - resized_item.size[0]) // 2
		item_y = y + (case_size - resized_item.size[1]) // 2

		# Paste RGBA image using itself as the mask (no split() and no extra mask object)
		showcase_image.paste(resized_item, (item_x, item_y), resized_item)

	# Save to output directory
	os.makedirs(output_dir, exist_ok=True)
	output_path = os.path.join(output_dir, filename)
	showcase_image = showcase_image.convert("RGB")
	showcase_image.save(output_path)





# ruff: noqa: E101
# Imports
from enum import Enum
from typing import cast

import stouputils as stp
from beet.core.utils import JsonDict

from ..__memory__ import Mem

# Constants
SLOTS: dict[str, str] = {
	"helmet": "head",
	"chestplate": "chest",
	"leggings": "legs",
	"boots": "feet",
	"sword": "mainhand",
	"pickaxe": "mainhand",
	"axe": "mainhand",
	"shovel": "mainhand",
	"hoe": "mainhand",
	"spear": "mainhand",
}
UNIQUE_SLOTS_VALUES: list[str] = []	# No sorted(set()) since we want to preserve the current order
for slot in SLOTS.values():
	if slot not in UNIQUE_SLOTS_VALUES:
		UNIQUE_SLOTS_VALUES.append(slot)

class DefaultOre(Enum):
	NETHERITE = "netherite"
	DIAMOND = "diamond"
	IRON = "iron"
	GOLD = "golden"
	CHAINMAIL = "stone"		# Stone tools
	COPPER = "copper"		# Copper added in 1.21.9
	LEATHER = "wooden"		# Wooden tools

class VanillaEquipments(Enum):
	""" Default vanilla equipments values (durability, armor, armor_toughness, knockback_resistance, attack_damage, attack_speed) """
	HELMET			= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 55,		"armor": 1},
						DefaultOre.COPPER:		{"durability": 121,		"armor": 2},
						DefaultOre.CHAINMAIL:	{"durability": 165,		"armor": 2},
						DefaultOre.IRON:		{"durability": 165,		"armor": 2},
						DefaultOre.GOLD:		{"durability": 77,		"armor": 2},
						DefaultOre.DIAMOND:		{"durability": 363,		"armor": 3,	"armor_toughness": 2},
			 			DefaultOre.NETHERITE:	{"durability": 407,		"armor": 3,	"armor_toughness": 3,	"knockback_resistance": 0.1}
					})
	CHESTPLATE		= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 80,		"armor": 3},
						DefaultOre.COPPER:		{"durability": 176,		"armor": 4},
						DefaultOre.CHAINMAIL:	{"durability": 240,		"armor": 5},
						DefaultOre.IRON:		{"durability": 240,		"armor": 6},
						DefaultOre.GOLD:		{"durability": 112,		"armor": 5},
						DefaultOre.DIAMOND: 	{"durability": 528,		"armor": 8,	"armor_toughness": 2},
						DefaultOre.NETHERITE:	{"durability": 592,		"armor": 8,	"armor_toughness": 3,	"knockback_resistance": 0.1}
					})
	LEGGINGS		= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 75,		"armor": 2},
						DefaultOre.COPPER:		{"durability": 165,		"armor": 3},
  						DefaultOre.CHAINMAIL:	{"durability": 225,		"armor": 4},
						DefaultOre.IRON:		{"durability": 225,		"armor": 5},
						DefaultOre.GOLD:		{"durability": 105,		"armor": 3},
						DefaultOre.DIAMOND:		{"durability": 495,		"armor": 6,	"armor_toughness": 2},
						DefaultOre.NETHERITE:	{"durability": 555,		"armor": 6,	"armor_toughness": 3,	"knockback_resistance": 0.1}
					})
	BOOTS			= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 65,		"armor": 1},
						DefaultOre.COPPER:		{"durability": 143,		"armor": 1},
						DefaultOre.CHAINMAIL:	{"durability": 195,		"armor": 1},
						DefaultOre.IRON:		{"durability": 195,		"armor": 2},
						DefaultOre.GOLD:		{"durability": 95,		"armor": 1},
						DefaultOre.DIAMOND:		{"durability": 429,		"armor": 3,	"armor_toughness": 2},
						DefaultOre.NETHERITE:	{"durability": 481,		"armor": 3,	"armor_toughness": 3,	"knockback_resistance": 0.1}
					})
	SWORD			= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 59,		"attack_damage": 4,		"attack_speed": -2.40},
						DefaultOre.COPPER:		{"durability": 190,		"attack_damage": 5,		"attack_speed": -2.40},
						DefaultOre.CHAINMAIL:	{"durability": 131,		"attack_damage": 5,		"attack_speed": -2.40},
						DefaultOre.IRON:		{"durability": 250,		"attack_damage": 6,		"attack_speed": -2.40},
						DefaultOre.GOLD:		{"durability": 32,		"attack_damage": 4,		"attack_speed": -2.40},
						DefaultOre.DIAMOND:		{"durability": 1561,	"attack_damage": 7,		"attack_speed": -2.40},
						DefaultOre.NETHERITE:	{"durability": 2031,	"attack_damage": 8,		"attack_speed": -2.40}
					})
	PICKAXE			= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 59,		"attack_damage": 2,		"attack_speed": -2.8},
						DefaultOre.COPPER:		{"durability": 190,		"attack_damage": 3,		"attack_speed": -2.8},
						DefaultOre.CHAINMAIL:	{"durability": 131,		"attack_damage": 3,		"attack_speed": -2.8},
						DefaultOre.IRON:		{"durability": 250,		"attack_damage": 4,		"attack_speed": -2.8},
						DefaultOre.GOLD:		{"durability": 32,		"attack_damage": 2,		"attack_speed": -2.8},
						DefaultOre.DIAMOND:		{"durability": 1561,	"attack_damage": 5,		"attack_speed": -2.8},
						DefaultOre.NETHERITE:	{"durability": 2031,	"attack_damage": 6,		"attack_speed": -2.8}
					})
	AXE				= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 59,		"attack_damage": 7,		"attack_speed": -3.20},
						DefaultOre.COPPER:		{"durability": 190,		"attack_damage": 9,		"attack_speed": -3.20},
						DefaultOre.CHAINMAIL:	{"durability": 131,		"attack_damage": 9,		"attack_speed": -3.20},
						DefaultOre.IRON:		{"durability": 250,		"attack_damage": 9,		"attack_speed": -3.10},
						DefaultOre.GOLD:		{"durability": 32,		"attack_damage": 7,		"attack_speed": -3.00},
						DefaultOre.DIAMOND:		{"durability": 1561,	"attack_damage": 9,		"attack_speed": -3.00},
						DefaultOre.NETHERITE:	{"durability": 2031,	"attack_damage": 10,	"attack_speed": -3.00}
					})
	SHOVEL			= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 59,		"attack_damage": 2.5,	"attack_speed": -3.00},
						DefaultOre.COPPER:		{"durability": 190,		"attack_damage": 3.5,	"attack_speed": -3.00},
						DefaultOre.CHAINMAIL:	{"durability": 131,		"attack_damage": 3.5,	"attack_speed": -3.00},
						DefaultOre.IRON:		{"durability": 250,		"attack_damage": 4.5,	"attack_speed": -3.00},
						DefaultOre.GOLD:		{"durability": 32,		"attack_damage": 2.5,	"attack_speed": -3.00},
						DefaultOre.DIAMOND:		{"durability": 1561,	"attack_damage": 5.5,	"attack_speed": -3.00},
			 			DefaultOre.NETHERITE:	{"durability": 2031,	"attack_damage": 6.5,	"attack_speed": -3.00}
					})
	HOE				= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 59,		"attack_damage": 1,		"attack_speed": -3.00},
						DefaultOre.COPPER:		{"durability": 190,		"attack_damage": 1,		"attack_speed": -2.00},
						DefaultOre.CHAINMAIL:	{"durability": 131,		"attack_damage": 1,		"attack_speed": -2.00},
						DefaultOre.IRON:		{"durability": 250,		"attack_damage": 1,		"attack_speed": -1.00},
						DefaultOre.GOLD:		{"durability": 32,		"attack_damage": 1,		"attack_speed": -3.00},
						DefaultOre.DIAMOND:		{"durability": 1561,	"attack_damage": 1,		"attack_speed": 0.00},
						DefaultOre.NETHERITE:	{"durability": 2031,	"attack_damage": 1,		"attack_speed": 0.00}
					})
	SPEAR			= cast(dict[DefaultOre, dict[str, float]],
					{	DefaultOre.LEATHER:		{"durability": 59,		"attack_damage": 1,		"attack_speed": -2.462},
						DefaultOre.COPPER:		{"durability": 190,		"attack_damage": 2,		"attack_speed": -2.824},
						DefaultOre.CHAINMAIL:	{"durability": 131,		"attack_damage": 2,		"attack_speed": -2.666},
						DefaultOre.IRON:		{"durability": 250,		"attack_damage": 3,		"attack_speed": -2.947},
						DefaultOre.GOLD:		{"durability": 32,		"attack_damage": 1,		"attack_speed": -2.947},
						DefaultOre.DIAMOND:		{"durability": 1561,	"attack_damage": 4,		"attack_speed": -3.048},
						DefaultOre.NETHERITE:	{"durability": 2031,	"attack_damage": 5,		"attack_speed": -3.130}
					})

class EquipmentsConfig:
	def __init__(self, equivalent_to: DefaultOre = DefaultOre.DIAMOND, pickaxe_durability: float | int = 0, attributes: dict[str, float] | None = None, ignore_recipes: bool = False) -> None:
		""" Creates a configuration for equipments (based on the pickaxe)

		Args:
			equivalent_to (DEFAULT_ORE):	The equivalent ore to compare to (ex: DEFAULT_ORE.DIAMOND)
			pickaxe_durability (int):		The pickaxe durability that will be used to calculate the durability of other equipments
			attributes (dict[str, float]):	(optional) Attributes with type "add_value" to add (not override) to the equipment (ex: "attack_damage": 1.0, means 6 attack damage for diamond pickaxe)
				{"attack_damage": 1.0, "armor": 1.0, "mining_efficiency": 1}
				attack_damage and mining_efficiency are always on tools
				armor and armor_toughness is always on armor
			ignore_recipes (bool):			Whether to ignore recipes generation for this material or not

		If you need a specific attribute for a generated item, you should append it afterward.
		"""
		if attributes is None:
			attributes = {}
		self.ignore_recipes: bool = ignore_recipes
		self.equivalent_to: DefaultOre = equivalent_to
		self.pickaxe_durability: int = int(pickaxe_durability if pickaxe_durability > 0 else VanillaEquipments.PICKAXE.value[equivalent_to]["durability"])
		self.attributes: dict[str, float] = attributes
		for key in attributes.keys():
			if "player." in key:
				stp.warning("Since 1.21.3, the 'player.' prefix is no longer written in attributes!!!")
			elif "generic." in key:
				stp.warning("Since 1.21.3, the 'generic.' prefix is no longer written in attributes!!!")
			if "knockback_resistance" in key and attributes[key] >= 1:
				stp.warning(f"You are setting the Knockback Resistance of an equipment to {attributes[key]}. Be aware that Minecraft automatically multiplies it by 10 when applied to an equipment.")

	def getter(self) -> tuple[DefaultOre, int, dict[str, float]]:
		return self.equivalent_to, self.pickaxe_durability, self.attributes

	@stp.simple_cache
	def get_tools_attributes(self) -> dict[str, float]:
		NOT_ON_TOOLS = ["armor", "armor_toughness", "knockback_resistance"]
		return {key: value for key, value in self.attributes.items() if key not in NOT_ON_TOOLS}

	@stp.simple_cache
	def get_armor_attributes(self) -> dict[str, float]:
		NOT_ON_ARMOR = ["attack_damage", "mining_efficiency"]
		return {key: value for key, value in self.attributes.items() if key not in NOT_ON_ARMOR}

	def __str__(self) -> str:
		return f"EquipmentsConfig(equivalent_to={self.equivalent_to}, pickaxe_durability={self.pickaxe_durability}, attributes={self.attributes}, ignore_recipes={self.ignore_recipes})"

	def __repr__(self) -> str:
		return self.__str__()

# UUIDs utils
def format_attributes(attributes: dict[str, float], slot: str, attr_config: dict[str, float] | None = None) -> list[JsonDict]:
	""" Returns generated attribute_modifiers key for an item (adds up attributes and config) """
	# Get attributes from config
	if attr_config is None:
		attr_config = {}
	attribute_modifiers: list[JsonDict] = []
	for attribute_name, value in attr_config.items():

		# We already have a base_attack_damage
		if attribute_name == "attack_damage":
			value -= 1

		# If not durability, we add the base attribute
		if attribute_name != "durability":
			if attribute_name in ["attack_damage", "attack_speed"]:
				attribute_modifiers.append({"type": attribute_name, "amount": value, "operation": "add_value", "slot": slot, "id": f"minecraft:base_{attribute_name}"})
			else:
				attribute_modifiers.append({"type": attribute_name, "amount": value, "operation": "add_value", "slot": slot, "id": f"{Mem.ctx.project_id}:{attribute_name}.{slot}"})

	# For each attribute, add it to the list if not in, else add the value
	for attribute_name, value in attributes.items():
		found = False
		for attribute in attribute_modifiers:
			if attribute["type"] == attribute_name:
				attribute["amount"] += value
				found = True
				break
		if not found:
			attribute_modifiers.append({"type": attribute_name, "amount": value, "operation": "add_value", "slot": slot, "id": f"{Mem.ctx.project_id}:{attribute_name}.{slot}"})

	# Return the list of attributes
	return attribute_modifiers





# Imports
import os
import shutil
from typing import cast

import requests
import stouputils as stp
from beet import Model
from beet.core.utils import JsonDict
from model_resolver.render import Render

from ...core.__memory__ import Mem
from ...core.cls.item import Item
from ...core.constants import (
	DOWNLOAD_VANILLA_ASSETS_RAW,
	DOWNLOAD_VANILLA_ASSETS_SOURCE,
	DOWNLOAD_VANILLA_ASSETS_SPECIAL_RAW,
)
from .shared_import import SharedMemory


# Generate iso renders for every item in the definitions
def generate_all_iso_renders():
	ns: str = Mem.ctx.project_id

	# Create the items folder
	path = SharedMemory.cache_path + "/items"
	os.makedirs(f"{path}/{ns}", exist_ok = True)

	# For every item, get the model path and the destination path
	cache_assets: bool = Mem.ctx.meta.get("stewbeet",{}).get("manual", {}).get("cache_assets", True)
	for_model_resolver: dict[str, str] = {}
	for item in Mem.definitions.keys():
		obj = Item.from_id(item)

		# Skip items that don't have models
		if not obj.components.get("item_model"):
			continue

		# Skip if item is already generated (to prevent OpenGL launching for nothing)
		if os.path.exists(f"{path}/{ns}/{item}.png") and cache_assets:
			continue

		# Add to the model resolver queue (only if present in resource pack)
		model: Model | None = Mem.ctx.assets[ns].models.get(f"item/{item}")
		rp_path = f"{ns}:item/{item}"
		dst_path = f"{path}/{ns}/{item}.png"
		if model is not None and model.get_content().get("textures", None) is not None: # type: ignore
			for_model_resolver[rp_path] = dst_path

	# Launch model resolvers for remaining blocks
	if len(for_model_resolver) > 0:

		## Model Resolver v0.12.0
		# model_resolver_main(
		# 	render_size = config['opengl_resolution'],
		# 	load_dir = load_dir,
		# 	output_dir = None,	# type: ignore
		# 	use_cache = False,
		# 	minecraft_version = "latest",
		# 	__special_filter__ = for_model_resolver	# type: ignore
		# )

		# If atlas is used in overlay, copy it
		any_atlas_used: bool = "before_format_73" in Mem.ctx.assets.overlays._wrapped.keys() # type: ignore
		if any_atlas_used:
			Mem.ctx.assets["minecraft"].atlases["temporary_stewbeet"] = Mem.ctx.assets.overlays["before_format_73"]["minecraft"].atlases["blocks"]

		## Model Resolver >= v1.12.0
		stp.debug(f"Generating iso renders for {len(for_model_resolver)} items, this may take a while...")
		render = Render(Mem.ctx)
		for rp_path, dst_path in for_model_resolver.items():
			render.add_model_task(rp_path, path_save=dst_path, animation_mode="one_file")
		render.run()
		stp.debug("Generated iso renders for all items")

		# Remove temporary atlas
		if any_atlas_used:
			del Mem.ctx.assets["minecraft"].atlases["temporary_stewbeet"]

	## Copy every used vanilla items
	# Get every used vanilla items
	used_vanilla_items: set[str] = set()
	for item in Mem.definitions.keys():
		obj = Item.from_id(item)
		for recipe in obj.recipes:
			ingredients = []
			if recipe.get("ingredients"):
				ingredients = recipe["ingredients"]
				if isinstance(ingredients, dict):
					ingredients = cast(list[JsonDict], ingredients.values())
			elif recipe.get("ingredient"):
				ingredients = [recipe["ingredient"]]
			for ingredient in ingredients:
				if "item" in ingredient:
					used_vanilla_items.add(ingredient["item"].split(":")[1])
			if recipe.get("result") and recipe["result"].get("item"):
				used_vanilla_items.add(recipe["result"]["item"].split(":")[1])
		pass

	# Download all the vanilla textures from the wiki
	def download_item(item: str, destination: str = ""):
		if not destination:
			destination = f"{path}/minecraft/{item}.png"
		if not (os.path.exists(destination) and cache_assets):	# If not downloaded yet or not using cache
			for base_link in (DOWNLOAD_VANILLA_ASSETS_SPECIAL_RAW, DOWNLOAD_VANILLA_ASSETS_RAW):
				for folder in ["item", "block", "items"]:
					link: str = f"{base_link}/{folder}/{item}.png"
					response = requests.get(link)
					if response.status_code == 200:
						with stp.super_open(destination, "wb") as file:
							return file.write(response.content)
			# If all attempts failed
			stp.warning(f"Failed to download texture for '{item}', please add it manually to '{destination}'")
			stp.warning(f"Suggestion link: '{DOWNLOAD_VANILLA_ASSETS_SOURCE}'")

	# Multithread the download
	stp.multithreading(download_item, used_vanilla_items, max_workers=min(32, len(used_vanilla_items)))

	# Download painting texture for custom paintings
	last_painting_path: str = ""
	for item, data in Mem.definitions.items():
		if data["id"] == "minecraft:painting" and not data.get("item_model"):
			if not last_painting_path:
				last_painting_path = f"{path}/{ns}/{item}.png"
				download_item("painting", last_painting_path)
			else:
				# Just copy the last painting downloaded
				shutil.copy(last_painting_path, f"{path}/{ns}/{item}.png")





# Imports
import stouputils as stp
from beet import Context
from beet.core.utils import JsonDict

from .constants import SorterFile
from .match import generate_sorter


# Main entry point
@stp.measure_time(message="Execution time of 'stewbeet.plugins.datapack.sorters'")
def beet_default(ctx: Context):
	""" Main entry point for the sorters plugin.
	This plugin generates functions that sort lists in storage.

	Args:
		ctx (Context): The beet context.
	"""
	ctx.data.extend_namespace.append(SorterFile)

	# Iterate through all sorter configurations
	for file_instance in ctx.data[SorterFile].values():
		sorter: JsonDict = file_instance.data.model_dump()
		generate_sorter(ctx, sorter)

	# Clear the registry
	ctx.data[SorterFile].clear()




import stouputils as stp
from ..__memory__ import Mem as Mem
from beet.core.utils import JsonDict as JsonDict
from enum import Enum

SLOTS: dict[str, str]
UNIQUE_SLOTS_VALUES: list[str]

class DefaultOre(Enum):
    NETHERITE = 'netherite'
    DIAMOND = 'diamond'
    IRON = 'iron'
    GOLD = 'golden'
    CHAINMAIL = 'stone'
    COPPER = 'copper'
    LEATHER = 'wooden'

class VanillaEquipments(Enum):
    """ Default vanilla equipments values (durability, armor, armor_toughness, knockback_resistance, attack_damage, attack_speed) """
    HELMET = ...
    CHESTPLATE = ...
    LEGGINGS = ...
    BOOTS = ...
    SWORD = ...
    PICKAXE = ...
    AXE = ...
    SHOVEL = ...
    HOE = ...
    SPEAR = ...

class EquipmentsConfig:
    ignore_recipes: bool
    equivalent_to: DefaultOre
    pickaxe_durability: int
    attributes: dict[str, float]
    def __init__(self, equivalent_to: DefaultOre = ..., pickaxe_durability: float | int = 0, attributes: dict[str, float] | None = None, ignore_recipes: bool = False) -> None:
        ''' Creates a configuration for equipments (based on the pickaxe)

\t\tArgs:
\t\t\tequivalent_to (DEFAULT_ORE):\tThe equivalent ore to compare to (ex: DEFAULT_ORE.DIAMOND)
\t\t\tpickaxe_durability (int):\t\tThe pickaxe durability that will be used to calculate the durability of other equipments
\t\t\tattributes (dict[str, float]):\t(optional) Attributes with type "add_value" to add (not override) to the equipment (ex: "attack_damage": 1.0, means 6 attack damage for diamond pickaxe)
\t\t\t\t{"attack_damage": 1.0, "armor": 1.0, "mining_efficiency": 1}
\t\t\t\tattack_damage and mining_efficiency are always on tools
\t\t\t\tarmor and armor_toughness is always on armor
\t\t\tignore_recipes (bool):\t\t\tWhether to ignore recipes generation for this material or not

\t\tIf you need a specific attribute for a generated item, you should append it afterward.
\t\t'''
    def getter(self) -> tuple[DefaultOre, int, dict[str, float]]: ...
    @stp.simple_cache
    def get_tools_attributes(self) -> dict[str, float]: ...
    @stp.simple_cache
    def get_armor_attributes(self) -> dict[str, float]: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...

def format_attributes(attributes: dict[str, float], slot: str, attr_config: dict[str, float] | None = None) -> list[JsonDict]:
    """ Returns generated attribute_modifiers key for an item (adds up attributes and config) """
